DrDamage = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "AceHook-2.1", "FuBarPlugin-2.0")
local L = AceLibrary("AceLocale-2.2"):new("DrDamage")
local BS = AceLibrary("Babble-Spell-2.2")
local SEA = AceLibrary("SpecialEvents-Aura-2.0")
local GT = AceLibrary("Gratuity-2.0")
local Deformat = AceLibrary("Deformat-2.0")

local type = type
local pairs = pairs
local tonumber = tonumber
local math_abs = math.abs
local math_floor = math.floor
local math_ceil = math.ceil

local defaults = {
	--Calculation
	HitCalc = false,
	HitTarget = false,
	TargetLevel = 3,
	TargetPlayer = false,
	SpellDamage = 0,
	Healing = 0,
	ShadowDamage = 0,
	FireDamage = 0,
	NatureDamage = 0,
	FrostDamage = 0,
	ArcaneDamage = 0,
	HolyDamage = 0,
	HitRating = 0,
	CritRating = 0,
	ManualDmg = false,
	ManualPlus = false,	
	--Actionbar:
	DisplayType = "Avg",
	FontSize = 8,
	FontXPosition = 0,
	FontYPosition = 0,
	FontColorDmg = { r = 1, g = 1, b = 0.2 },
	FontColorHeal = { r = 0.4, g = 1.0, b = 0.3 },
	CastsLeft = false,
	CastsLeftDmg = false,
	UpdateShift = false,
	UpdateAlt = false,
	UpdateCtrl = false,
	BlizzardAB = false,
	--Tooltip:
	Display = true,
	DefaultColor = false,
	PlusDmg = true,
	Coeffs = true,
	DispCrit = true,
	DispHit = true,
	AvgHit = true,
	AvgCrit = true,
	Ticks = true,
	Total = true,
	Extra = true,
	Next = true,
	DPS = true,
	DPM = true,
	Doom = true,
	Casts = true,
	TwoRoll = false,
	ManaUsage = false,
}

DrDamage:RegisterDB("DrDamageDB");
DrDamage:RegisterDefaults("profile", defaults )

function DrDamage:OnInitialize()
	if IsAddOnLoaded( "FlexBar2" ) then
		FlexBar2:RegisterTextSub("drd", "DrDamage_Update",
		function( Button, Frame, updateSpells )
		
			if Button:GetEffectiveAttribute('type') == 'action' then
			
				local id = Button:GetEffectiveAttribute( "action" )
				
				if id then
					--[[
					if updateSpells then

						local gtype, pid = GetActionInfo( id )

						if gtype == "spell" then --if gtype =="macro"?
							local spellName = GetSpellName( pid, BOOKTYPE_SPELL );	

							if DrDamage:MatchData( updateSpells, spellName ) then
								return DrDamage:CheckAction( id )
							end
						end
					else
					--]]
						return DrDamage:CheckAction( id )
					--end
				end
				
				return false
			end
		end)
	end
	if IsAddOnLoaded( "InfiniBar" ) then
		InfiniBar:RegisterTextSub("drd", "DrDamage_Update",
		function( bar, btn, options, updateSpells )
			local spellName, Rank = bar:GetSpellNameAndRank( btn )
			if spellName then
				if not updateSpells or DrDamage:MatchData( updateSpells, spellName ) then
					return DrDamage:CheckAction( nil, nil, nil, nil, spellName, Rank )
				end
			end
			
			return false
		end)
	end
end

local changesDMG = {
    "CHARACTER_POINTS_CHANGED",
    "PLAYER_ALIVE",
}

function DrDamage:CHARACTER_POINTS_CHANGED()
    	DrDamage:DmgUpdateTalents()
end

function DrDamage:PLAYER_ALIVE()
    	DrDamage:DmgUpdateTalents()
end

local changesAB = {
    "ACTIONBAR_PAGE_CHANGED",
    "ACTIONBAR_HIDEGRID",
    "UPDATE_SHAPESHIFT_FORM",
    "UNIT_INVENTORY_CHANGED",
    "PLAYER_TARGET_CHANGED",
    "UPDATE_MACROS",
    "UNIT_MANA",
}

local _, englishClass = UnitClass("player")
local Mana_Cost
local visualChange = true
local previousValues = {
	GetSpellBonusDamage(1),
	GetSpellBonusDamage(2),
	GetSpellBonusDamage(3),
	GetSpellBonusDamage(4),
	GetSpellBonusDamage(5),
	GetSpellBonusDamage(6),
	GetSpellBonusDamage(7),
	GetSpellBonusHealing(),
	GetCombatRatingBonus(8),
	GetSpellCritChance(1),
}

local bAP, pAP, nAP = UnitAttackPower( "player" )


function DrDamage:ACTIONBAR_PAGE_CHANGED()

	if not self:IsEventScheduled("UpdatingAB") then
		self:ScheduleEvent("UpdatingAB", function() DrDamage:UpdateAB( false, true ) end, 0.5, self)
		visualChange = true
	end
end

function DrDamage:ACTIONBAR_HIDEGRID()

	if not self:IsEventScheduled("UpdatingAB") and ( GetCursorInfo() == "spell" or GetCursorInfo() == "macro" )  then
		self:ScheduleEvent("UpdatingAB", function() DrDamage:UpdateAB( false, true ) end, 0.5, self)
		visualChange = true
	end
end

function DrDamage:UPDATE_SHAPESHIFT_FORM()

	if self:IsEventScheduled("UpdatingAB") then
		 self:CancelScheduledEvent("UpdatingAB")
	end
		 
	self:ScheduleEvent("UpdatingAB", function() DrDamage:UpdateAB( false, true ) end, 0.5, self)
	visualChange = true
end

local oldBuffed
local oldDebuffed
local oldTargetType

function DrDamage:PLAYER_TARGET_CHANGED()

	if not self:IsEventScheduled("UpdatingAB") then
	
		local buffed = false
		local debuffed = false
		local targettype = false
		
		if DrDamage.Debuffs then
		
			if not oldDebuffed and not oldBuffed and not DrDamage.db.profile.HitCalc then

				for debuffName in SEA:DebuffIter( "target" ) do

					if BS:HasReverseTranslation( debuffName ) then
						debuffName = BS:GetReverseTranslation( debuffName )
					end	

					if DrDamage.Debuffs[debuffName] then
						debuffed = true
					end
				end

				if not debuffed then
					for buffName in SEA:BuffIter( "target" ) do

						if BS:HasReverseTranslation( buffName ) then
							buffName = BS:GetReverseTranslation( buffName )
						end

						if DrDamage.HealingBuffs[buffName] then
							buffed = true
						end
					end
				end
			end	
		end
		
		if UnitCreatureType( "target" ) ~= oldTargetType then
			targettype = true
			oldTargetType = UnitCreatureType( "target" )
		end
		
		if englishClass == "SHAMAN" then
			targettype = true
		end
			
		if targettype or buffed or debuffed or oldDebuffed or oldBuffed or DrDamage.db.profile.HitCalc then

			self:ScheduleEvent("UpdatingAB", self.UpdateAB, 2, self)

			if debuffed then
				oldDebuffed = true
			else
				oldDebuffed = false
			end
			
			if buffed then
				oldBuffed = true
			else
				oldBuffed = false
			end
		end
	end
end

function DrDamage:UPDATE_MACROS()

	if not self:IsEventScheduled("UpdatingAB") then
		self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1, self)
		visualChange = true
	end
end

local playerMana = UnitManaMax( "player" )

function DrDamage:UNIT_MANA( unit )

	if not DrDamage.db.profile.CastsLeft and not DrDamage.db.profile.CastsLeftDmg then
		do return end
	end
	
	if unit == "player" then
		
		local update = false
		local manaD = math_abs( UnitMana( "player" ) - playerMana )
	
		if manaD and manaD >= 200 then
			update = true
			playerMana = UnitMana( "player" )
		end
		
		if update and not self:IsEventScheduled( "UpdatingAB" ) then
			self:ScheduleEvent("UpdatingAB", function() DrDamage:UpdateAB( false, false, true ) end, 1, self)
		end
	end
end

function DrDamage:UNIT_INVENTORY_CHANGED( unit )

	if unit == "player" then
	
		local update = false
		
		for i = 1, 7 do
			if GetSpellBonusDamage(i) ~= previousValues[i] then
				previousValues[i] = GetSpellBonusDamage(i)
				update = true
			end
		end
		
		if GetSpellBonusHealing() ~= previousValues[8] then
			previousValues[8] = GetSpellBonusHealing()
			update = true
		end
		
		if GetCombatRatingBonus(8) ~= previousValues[9] then
			previousValues[9] = GetCombatRatingBonus(8)
			update = true
		end
		
		if DrDamage:SpellCrit() ~= previousValues[10] then
			previousValues[10] = DrDamage:SpellCrit()
			update = true
		end
		
		local rangedItem = GetInventoryItemLink( "player", GetInventorySlotInfo( "RangedSlot" ) )
		
		if rangedItem and GetItemInfo( rangedItem ) ~= previousValues[11] or not rangedItem and previousValues[11] then
			if rangedItem then
				previousValues[11] = GetItemInfo( rangedItem )
			else
				previousValues[11] = nil
			end
			
			update = true
		end
		
		if englishClass == "PALADIN" then
			local nbAP, npAP, nnAP = UnitAttackPower( "player" )
			
			if bAP and pAP and nAP and nbAP and bpAP and nnAP and ( bAP + pAP + nAP ) ~= ( nbAP + bpAP + nnAP ) then
				update = true
				bAP = nbAP
				pAP = npAP
				nAP = nnAP
			end
		end
	
		if update then

			if type( DrDamage.ItemSets ) == "function" then
				DrDamage:ItemSets()
			end
		
			DrDamage:UpdateAB()			
		end
	end
end

local updateABSpell
local updatingSpell

function DrDamage:PlayerAuraUpdate( buffName )

	if self:IsEventScheduled("UpdatingAB") and not updateABSpell then
		do return end
	end
	
	local update = false
	
	for i = 1, 7 do
		if GetSpellBonusDamage(i) ~= previousValues[i] then
			previousValues[i] = GetSpellBonusDamage(i)
			update = true
		end
	end

	if GetSpellBonusHealing() ~= previousValues[8] then
		previousValues[8] = GetSpellBonusHealing()
		update = true
	end

	if GetCombatRatingBonus(8) ~= previousValues[9] then
		previousValues[9] = GetCombatRatingBonus(8)
		update = true
	end

	if DrDamage:SpellCrit() ~= previousValues[10] then
		previousValues[10] = DrDamage:SpellCrit()
		update = true
	end
		
	if update then
		DrDamage:UpdateAB()
		do return end
	end
	
	if not DrDamage.DamageBuffs then
		do return end
	end
	
	local updateSpell
	
	if englishClass == "PALADIN" then
	
		local nbAP, npAP, nnAP = UnitAttackPower( "player" )

		if bAP and pAP and nAP and nbAP and bpAP and nnAP and ( bAP + pAP + nAP ) ~= ( nbAP + bpAP + nnAP ) then
			update = true
			updateSpell = BS["Crusader Strike"]
			bAP = nbAP
			pAP = npAP
			nAP = nnAP
		end
	end	
	
	if BS:HasReverseTranslation( buffName ) then
		buffName = BS:GetReverseTranslation( buffName )
	end
			
	if DrDamage.DamageBuffs[buffName] then
		updateSpell = DrDamage.DamageBuffs[buffName].Spell
		update = true
	elseif DrDamage.DamageDebuffs[buffName] then
		updateSpell = DrDamage.DamageDebuffs[buffName].Spell
		update = true
	end
	
	if self:IsEventScheduled("UpdatingAB") and updateABSpell then
		if updatingSpell == updateSpell then
			updateABSpell = false
		end
	end	

	if update then
		if updateSpell and not updateABSpell then
			self:ScheduleEvent("UpdatingAB", function() DrDamage:UpdateAB( false, false, false, updateSpell ) end, 0.5, self)
			updateABSpell = true
			updatingSpell = updateSpell
		else
			self:ScheduleEvent("UpdatingAB", self.UpdateAB, 0.5, self)
			updateABSpell = false
		end
	end
end


function DrDamage:TargetAuraUpdate( unitID, buffName )

	if not DrDamage.HealingBuffs or not DrDamage.Debuffs or self:IsEventScheduled("UpdatingAB") and not updateABSpell then
		do return end
	end
	
	if unitID == "target" then
	
		local update = false
		local updateSpell
		local healer = false
		
		if englishClass == "PRIEST" or englishClass == "SHAMAN" or englishClass == "PALADIN" or englishClass == "DRUID" then
			healer = true
		end
		
		if BS:HasReverseTranslation( buffName ) then
			buffName = BS:GetReverseTranslation( buffName )
		end		
		
		--|| Update healing auras only for healers ||
		if healer then
			if DrDamage.HealingBuffs[buffName] then

				if DrDamage.HealingBuffs[buffName].Class then
					for i, class in ipairs( DrDamage.HealingBuffs[buffName].Class ) do
						if class == englishClass then
							update = true
						end
					end
				elseif healer then
					update = true
				end
				
				if update then
					updateSpell = DrDamage.HealingBuffs[buffName].Spell
				end
			end
		end

		if not update then
		
			if DrDamage.Debuffs[buffName] then
				
				local debuff = DrDamage.Debuffs[buffName]
				
				if debuff.Class then
					for _, class in ipairs( debuff.Class ) do
						if class == englishClass then
							update = true
						end
					end				
				elseif debuff.School == "Healing" then
					if healer then
						update = true
					end
				elseif debuff.Affliction then
					if englishClass == "WARLOCK" then
						update = true
					end
				elseif debuff.Stun then
					if englishClass == "PALADIN" then
						update = true
					end
				else
					update = true
				end
				
				if update then
					updateSpell = DrDamage.Debuffs[buffName].Spell
				end
			end
		end
		
		if self:IsEventScheduled("UpdatingAB") and updateABSpell then
			if updatingSpell == updateSpell then
				updateABSpell = false
			end
		end
		
		if update then
			if updateSpell and not updateABSpell then
				self:ScheduleEvent("UpdatingAB", function() DrDamage:UpdateAB( false, false, false, updateSpell ) end, 0.75, self )
				updateABSpell = true
				updatingSpell = updateSpell				
			else
				self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1.5, self)
				updateABSpell = false
			end
		end
	end
end

local talents = {}
local spellInfo = {}
local talentEffects = {}
local _G

function DrDamage:OnEnable()

	if not DrDamage:DetermineClass() then
		do return end
	end
	
	Mana_Cost = MANA_COST
	talents = {}
	_G = getfenv(0)
	
	DrDamage:Hook( GameTooltip, "SetAction", true)
	DrDamage:SecureHook( GameTooltip, "SetSpell" )
	DrDamage:DmgUpdateTalents()
	
	if type( DrDamage.ItemSets ) == "function" then
		DrDamage:ItemSets()
	end
		
    	for i = 1, #changesDMG do
    		
    		if changesDMG[i] then
    			self:RegisterEvent( changesDMG[i] )
    		end
  	end
  	
  	DrDamage:DetermineAB()
	
	if DrDamage.db.profile.DisplayType ~= "Off" then
    		
    		for i = 1, #changesAB do
    	
    			if changesAB[i] then
    				self:RegisterEvent( changesAB[i] )
    			end
    		end
    		
    		self:RegisterEvent( "SpecialEvents_PlayerBuffGained", "PlayerAuraUpdate" )
    		self:RegisterEvent( "SpecialEvents_PlayerBuffCountChanged", "PlayerAuraUpdate" )
    		self:RegisterEvent( "SpecialEvents_PlayerBuffLost", "PlayerAuraUpdate" )
    		self:RegisterEvent( "SpecialEvents_PlayerDebuffGained", "PlayerAuraUpdate" )
    		self:RegisterEvent( "SpecialEvents_PlayerDebuffCountChanged", "PlayerAuraUpdate" )
    		self:RegisterEvent( "SpecialEvents_PlayerDebuffLost", "PlayerAuraUpdate" )
    		self:RegisterEvent( "SpecialEvents_UnitBuffGained", "TargetAuraUpdate" )
    		self:RegisterEvent( "SpecialEvents_UnitBuffCountChanged", "TargetAuraUpdate" )
    		self:RegisterEvent( "SpecialEvents_UnitBuffLost", "TargetAuraUpdate" )
    		self:RegisterEvent( "SpecialEvents_UnitDebuffGained", "TargetAuraUpdate" )
    		self:RegisterEvent( "SpecialEvents_UnitDebuffCountChanged", "TargetAuraUpdate" )
    		self:RegisterEvent( "SpecialEvents_UnitDebuffLost", "TargetAuraUpdate" ) 
  	end  	
end

function DrDamage:OnDisable()

	DrDamage:UpdateAB( true )
	talents = nil
	spellInfo = nil
	talentEffects = nil
	_G = nil
	
end

function DrDamage:dGet( n )
	return function() return DrDamage.db.profile[n] 
	end
end

function DrDamage:dSet( n, v, noUpd )
	return function(v) 
	
		DrDamage.db.profile[n] = v
		
		if not noUpd and not self:IsEventScheduled("UpdatingAB") then
			self:ScheduleEvent("UpdatingAB", function() DrDamage:UpdateAB( false, true ) end, 1.0, self)
		end	
	end
end

local DisplayTypes = { "Avg", "DPS", "DPSC", "Off", "Min", "Max", "AvgHit" }

--LOCALE:
local options = { 
	type='group',
	args = {
		ActionBar = {
			type = "group", desc = L["Options for the actionbar"], name = L["ActionBar"],
			args = {
				DisplayType = {
					name = L["Display"],
					desc = L["Choose what to display on the actionbar"],
					type = "text",
					validate = DisplayTypes,
					get =  DrDamage:dGet("DisplayType"),
					set =  DrDamage:dSet("DisplayType", value),
					order = 70,
				}, 				
				FontSize = {
					type = 'range',
					name = L["Actionbar font size"],
					desc = L["Set the font size"],
					min = 6,
					max = 20,
					step = 1,
					order = 75,
					get =  DrDamage:dGet("FontSize"),
					set = function( v ) 
						DrDamage.db.profile.FontSize = v
						visualChange = true
						DrDamage:UpdateAB()
					end,
				},
				FontXPosition = {
					type = 'range',
					name = L["Actionbar font horizontal position"],
					desc = L["Set the font horizontal position"],
					min = -5,
					max = 5,
					step = 1,
					order = 80,
					get =  DrDamage:dGet("FontXPosition"),
					set = function( v ) 
						DrDamage.db.profile.FontXPosition = v
						visualChange = true
						DrDamage:UpdateAB()
					end,
				},
				FontYPosition = {
					type = 'range',
					name = L["Actionbar font vertical position"],
					desc = L["Set the font vertical position"],
					min = -15,
					max = 30,
					step = 1,
					order = 85,
					get =  DrDamage:dGet("FontYPosition"),
					set = function( v ) 
						DrDamage.db.profile.FontYPosition = v
						visualChange = true
						DrDamage:UpdateAB()
					end,
				},
				FontColorDmg = {
					type = "text",
					name = L["Damage text color"],
					desc = L["Set the actionbar damage text color"],
					usage = L["<r=n, g=n, b=n> where n is 0-1"],
					order = 90,
					get = function()
						return "r="..DrDamage.db.profile.FontColorDmg.r..", g="..DrDamage.db.profile.FontColorDmg.g..", b="..DrDamage.db.profile.FontColorDmg.b
					end,
					set = function( colors )
					
						for k, v in string.gmatch( string.lower( colors ), "([r|g|b])=([%d\.]+)" ) do
						
							local error = false
						
							if v and ( tonumber( v ) > 1 or tonumber( v ) < 0  ) or not v then
								error = true
							end
							
							if not error then
								if k == "r" then
									DrDamage.db.profile.FontColorDmg.r = v
								elseif k == "g" then
									DrDamage.db.profile.FontColorDmg.g = v
								elseif k == "b" then
									DrDamage.db.profile.FontColorDmg.b = v
								end
							end
						end
						
						visualChange = true
						DrDamage:UpdateAB()
					end
				},
				FontColorHeal = {
					type = "text",
					name = L["Heal text color"],
					desc = L["Set the actionbar heal text color"],
					usage = L["<r=n, g=n, b=n> where n is 0-1"],
					order = 95,
					get = function()
						return "r="..DrDamage.db.profile.FontColorHeal.r..", g="..DrDamage.db.profile.FontColorHeal.g..", b="..DrDamage.db.profile.FontColorHeal.b
					end,
					set = function( colors )
					
						for k, v in string.gmatch( string.lower( colors ), "([r|g|b])=([%d\.]+)" ) do
						
							local error = false
						
							if v and ( tonumber( v ) > 1 or tonumber( v ) < 0  ) or not v then
								error = true
							end
							
							if not error then
								if k == "r" then
									DrDamage.db.profile.FontColorHeal.r = v
								elseif k == "g" then
									DrDamage.db.profile.FontColorHeal.g = v
								elseif k == "b" then
									DrDamage.db.profile.FontColorHeal.b = v
								end
							end
						end
						
						visualChange = true 
						DrDamage:UpdateAB()
					end
				},				
				CastsLeftDmg = {
					type = 'toggle',
					name = L["Damage spell casts left"],
					desc = L["Toggles amount of casts left to display on actionbar for damaging spells"],
					order = 100,
					get = DrDamage:dGet( "CastsLeftDmg" ),
					set = DrDamage:dSet( "CastsLeftDmg", value ),
				},
				CastsLeft = {
					type = 'toggle',
					name = L["Healing casts left"],
					desc = L["Toggles amount of casts left to display on actionbar for heals"],
					order = 105,
					get = DrDamage:dGet( "CastsLeft" ),
					set = DrDamage:dSet( "CastsLeft", value ),
				},				
        			Update = {
            				name = L["Update"],
            				desc = L["Forces update to actionbar"],
            				type = "execute",
            				order = 110,
            				func = "UpdateAB",
    				},
				UpdateShift = {
					type = 'toggle',
					name = L["Updates BT3 actionbars on shift modifier key."],
					desc = L["Toggles updates on modifier key on/off. Requires Reloading UI."],
					order = 115,
					get =  DrDamage:dGet( "UpdateShift" ),
					set =  DrDamage:dSet( "UpdateShift", value, true ),				
				},
				UpdateAlt = {
					type = 'toggle',
					name = L["Updates BT3 actionbars on alt modifier key."],
					desc = L["Toggles updates on modifier key on/off. Requires Reloading UI."],
					order = 120,
					get =  DrDamage:dGet( "UpdateAlt" ),
					set =  DrDamage:dSet( "UpdateAlt", value, true ),				
				}, 
				UpdateCtrl = {
					type = 'toggle',
					name = L["Updates BT3 actionbars on ctrl modifier key."],
					desc = L["Toggles updates on modifier key on/off. Requires Reloading UI."],
					order = 125,
					get =  DrDamage:dGet( "UpdateCtrl" ),
					set =  DrDamage:dSet( "UpdateCtrl", value, true ),				
				}, 
				BlizzardAB = {
					type = 'toggle',
					name = L["Support for Blizzard AB even with other AB addons running."],
					desc = L["Toggles updating on for the default actionbar."],
					order = 130,
					get =  DrDamage:dGet( "BlizzardAB" ),
					set =  DrDamage:dSet( "BlizzardAB", value ),				
				},				
			},
		},
		Tooltip = {
			type = "group", desc = L["Options for the tooltips"], name = L["Tooltip"],
			args = {
				Display = {
					type = 'toggle',
					name = L["Tooltip on/off"],
					desc = L["Toggles the tooltip info on/off"],
					order = 50,
					get = DrDamage:dGet( "Display" ),
					set = DrDamage:dSet( "Display", value, true ),
				},
				DefaultColor = {
					type = 'toggle',
					name = L["Default tooltip colors"],
					desc = L["Toggles the default blizzard tooltip colors on/off"],
					order = 51,
					get = DrDamage:dGet( "DefaultColor" ),
					set = DrDamage:dSet( "DefaultColor", value, true ),
				},					
				PlusDmg = {
					type = 'toggle',
					name = L["Show efficient +dmg/+healing"],
					desc = L["Toggles displaying of efficient +dmg/+healing"],
					order = 52,
					get = DrDamage:dGet( "PlusDmg" ),
					set = DrDamage:dSet( "PlusDmg", value, true ),
				},
				Coeffs = {
					type = 'toggle',
					name = L["Show coefficients"],
					desc = L["Toggles displaying of spell calculation data"],
					order = 53,
					get = DrDamage:dGet( "Coeffs" ),
					set = DrDamage:dSet( "Coeffs", value, true ),
				},				
				DispCrit = {
					type = 'toggle',
					name = L["Show crit %"],
					desc = L["Toggles displaying of crit %"],
					order = 54,
					get = DrDamage:dGet( "DispCrit" ),
					set = DrDamage:dSet( "DispCrit", value, true ),
				},
				DispHit = {
					type = 'toggle',
					name = L["Show hit %"],
					desc = L["Toggles displaying of hit %."],
					order = 56,
					get = DrDamage:dGet( "DispHit" ),
					set = DrDamage:dSet( "DispHit", value, true ),
				},				
				AvgHit = {
					type = 'toggle',
					name = L["Show avg + hit range"],
					desc = L["Toggles displaying of avg hit"],
					order = 58,
					get = DrDamage:dGet( "AvgHit" ),
					set = DrDamage:dSet( "AvgHit", value, true ),
				},
				AvgCrit = {
					type = 'toggle',
					name = L["Show avg crit + crit range"],
					desc = L["Toggles displaying of avg crit"],
					order = 60,
					get = DrDamage:dGet( "AvgCrit" ),
					set = DrDamage:dSet( "AvgCrit", value, true ),
				},
				Ticks = {
					type = 'toggle',
					name = L["Show per tick/hit"],
					desc = L["Toggles displaying of per hit/tick values"],
					order = 62,
					get = DrDamage:dGet( "Ticks" ),
					set = DrDamage:dSet( "Ticks", value, true ),
				},
				Total = {
					type = 'toggle',
					name = L["Show avg total dmg/heal"],
					desc = L["Toggles displaying of average in total values"],
					order = 64,
					get = DrDamage:dGet( "Total" ),
					set = DrDamage:dSet( "Total", value, true ),
				},				
				Extra = {
					type = 'toggle',
					name = L["Show extra dmg/heal (DoT, Ignite, Chained)"],
					desc = L["Toggles displaying of extra info"],
					order = 66,
					get = DrDamage:dGet( "Extra" ),
					set = DrDamage:dSet( "Extra", value, true ),
				},
				Next = {
					type = 'toggle',
					name = L["Show next values (+1% crit, +10 dmg)"],
					desc = L["Toggles displaying of +1% crit, +10 dmg values"],
					order = 68,
					get = DrDamage:dGet( "Next" ),
					set = DrDamage:dSet( "Next", value, true ),
				},
				DPS = {
					type = 'toggle',
					name = L["Show DPS/HPS"],
					desc = L["Toggles displaying of DPS/HPS"],
					order = 70,
					get = DrDamage:dGet( "DPS" ),
					set = DrDamage:dSet( "DPS", value, true ),
				},
				DPM = {
					type = 'toggle',
					name = L["Show DPM/HPM"],
					desc = L["Toggles displaying of DPM/HPM"],
					order = 72,
					get = DrDamage:dGet( "DPM" ),
					set = DrDamage:dSet( "DPM", value, true ),
				},
				Doom = {
					type = 'toggle',
					name = L["Show Damage/healing until OOM"],
					desc = L["Toggles displaying of damage/healing until OOM"],
					order = 74,
					get = DrDamage:dGet( "Doom" ),
					set = DrDamage:dSet( "Doom", value, true ),
				},
				Casts = {
					type = 'toggle',
					name = L["Show casts and time until OOM."],
					desc = L["Toggles displaying of casts and time until OOM (regen included)"],
					order = 75,
					get = DrDamage:dGet( "Casts" ),
					set = DrDamage:dSet( "Casts", value, true ),
				},
				ManaUsage = {
					type = 'toggle',
					name = L["Show additional mana usage information."],
					desc = L["Toggles displaying of true mana cost (if different) and mana per seconds cast"],
					order = 76,
					get = DrDamage:dGet( "ManaUsage" ),
					set = DrDamage:dSet( "ManaUsage", value, true ),
				},				
			},
		},		
		Calculation = {
			type = "group", desc = L["Calculation options"], name = L["Calculation"],
			args = {
				TwoRoll = {
					type = 'toggle',
					name = L["Toggle usage of two roll hit calculation"],
					desc = L["Toggles two roll hit calculation on. Don't change this unless you know what this does."],
					order = 64,
					get = DrDamage:dGet( "TwoRoll" ),
					set = DrDamage:dSet( "TwoRoll", value ),
				},			
				HitCalc = {			
					type = 'toggle',
					name = L["+Hit calculation"],
					desc = L["Toggles +hit calculation effects into averages on/off"],
					order = 65,
					get = DrDamage:dGet( "HitCalc" ),
					set = DrDamage:dSet( "HitCalc", value ),
				},
				HitTarget = {			
					type = 'toggle',
					name = L["Hit calculation by target level"],
					desc = L["Toggles +hit calculation by target level."],
					order = 66,
					get = DrDamage:dGet( "HitTarget" ),
					set = DrDamage:dSet( "HitTarget", value ),
				},
				TargetLevel = {
					type = 'range',
					name = L["Manual target level setting"],
					desc = L["Manual set of target level compared to your level. Make sure 'HitTarget' is turned off to use this."],
					min = -10,
					max = 10,
					step = 1,
					order = 67,
					get = DrDamage:dGet( "TargetLevel" ),
					set = DrDamage:dSet( "TargetLevel", value ),
				},
				TargetPlayer = {
					type = 'toggle',
					name = L["Manual target is a player."],
					desc = L["Toggles if your manually set target is a player."],
					order = 68,
					get = DrDamage:dGet( "TargetPlayer" ),
					set = DrDamage:dSet( "TargetPlayer", value ),
				},				
				ManualDmg = {
					type = 'toggle',
					name = L["Manual variables"],
					desc = L["Allows you to manually set damage/healing properties"],
					order = 71,
					get = DrDamage:dGet( "ManualDmg" ),
					set = DrDamage:dSet( "ManualDmg", value ),
				},
				ManualPlus = {
					type = 'toggle',
					name = L["Manual damage adds to detected"],
					desc = L["Allows you to set if manual values should add to detected values"],
					order = 72,
					get = DrDamage:dGet( "ManualPlus" ),
					set = DrDamage:dSet( "ManualPlus", value ),
				}, 
				Healing = {
					type = 'range',
					name = L["Healing"],
					min = -500,
					max = 2000,
					step = 1,
					desc = L["Input your +healing"],
					order = 79,
					get = DrDamage:dGet("Healing"),
					set = DrDamage:dSet("Healing", value),
				},				
				SpellDamage = {
					type = 'range',
					name = L["Spell Damage"],
					min = -500,
					max = 2000,
					step = 1,
					desc = L["Input your spell damage"],
					order = 80,
					get = DrDamage:dGet("SpellDamage"),
					set = DrDamage:dSet("SpellDamage", value),
				},		
				FireDamage = {
					type = 'range',
					name = L["Fire Damage"],
					min = -500,
					max = 2000,
					step = 1,
					desc = L["Input your fire +damage"],
					order = 81,
					get = DrDamage:dGet("FireDamage"),
					set = DrDamage:dSet("FireDamage", value),		
		
				},
				ShadowDamage = {
					type = 'range',
					name = L["Shadow Damage"],
					min = -500,
					max = 2000,
					step = 1,
					desc = L["Input your shadow +damage"],
					order = 82,
					get = DrDamage:dGet("ShadowDamage"),
					set = DrDamage:dSet("ShadowDamage", value),		
				},
				FrostDamage = {
					type = 'range',
					name = L["Frost Damage"],
					min = -500,
					max = 2000,
					step = 1,
					desc = L["Input your frost +damage"],
					order = 83,
					get = DrDamage:dGet("FrostDamage"),
					set = DrDamage:dSet("FrostDamage", value),		
				},
				NatureDamage = {
					type = 'range',
					name = L["Nature Damage"],
					min = -500,
					max = 2000,
					step = 1,
					desc = L["Input your nature +damage"],
					order = 84,
					get = DrDamage:dGet("NatureDamage"),
					set = DrDamage:dSet("NatureDamage", value),		
				},
				ArcaneDamage = {
					type = 'range',
					name = L["Arcane Damage"],
					min = -500,
					max = 2000,
					step = 1,
					desc = L["Input your arcane +damage"],
					order = 85,
					get = DrDamage:dGet("ArcaneDamage"),
					set = DrDamage:dSet("ArcaneDamage", value),		
				},
				HolyDamage = {
					type = 'range',
					name = L["Holy Damage"],
					min = -500,
					max = 2000,
					step = 1,
					desc = L["Input your holy +damage"],
					order = 85,
					get = DrDamage:dGet("HolyDamage"),
					set = DrDamage:dSet("HolyDamage", value),		
				},					
				CritRating = {
					type = 'range',
					name = L["Manual critical rating"],
					min = -400,
					max = 1000,
					step = 1,
					desc = L["Input critical rating you want to use."],
					order = 88,
					get = DrDamage:dGet("CritRating"),
					set = DrDamage:dSet("CritRating", value),					
				},
				HitRating = {
					type = 'range',
					name = L["Manual hit rating"],
					min = -200,
					max = 500,
					step = 1,
					desc = L["Input hit rating you want to use."],
					order = 89,
					get = DrDamage:dGet("HitRating"),
					set = DrDamage:dSet("HitRating", value),	
				},
			},
		},
	},
};

DrDamage:RegisterChatCommand( { "/drdmg", "/drdamage" }, options)

--FuBar options
DrDamage.name = "DrDamage"
DrDamage.title = "Dr. Damage"
DrDamage.hasNoColor = true
DrDamage.hasNoText = true
DrDamage.cannotDetachTooltip = true
DrDamage.OnMenuRequest = options
DrDamage.hideWithoutStandby = true
DrDamage.hasIcon = "Interface\\Icons\\Spell_Holy_SearingLightPriest"

function DrDamage:ClearTable( table )
	for k in pairs( table ) do
		table[k] = nil
	end
end

function DrDamage:DmgUpdateTalents()

	DrDamage:ClearTable( talents )
	
	for t = 1, GetNumTalentTabs() do
	
    		local numTalents = GetNumTalents( t )
    		  		
	    	for i = 1, numTalents do
	    	
	    		local nameTalent, _, _, _, currRank = GetTalentInfo( t, i )
	    		
	    		if BS:HasReverseTranslation( nameTalent ) then
	    			nameTalent = BS:GetReverseTranslation( nameTalent )
	    		end
	    		
	    	    	if currRank ~= 0 then
	    	    			    	    
	    	    		if talentEffects[nameTalent] then
	    	    		
	    	    			if talentEffects[nameTalent].Locale then
	    	    				if DrDamage:MatchData( talentEffects[nameTalent].Locale, GetLocale() ) then
	    	    					talents[nameTalent] = currRank
	    	    				end
	    	    			else
	    	    				talents[nameTalent] = currRank
	    	    			end
	    	    			--self:Print("Talent name: " .. nameTalent .. " Points: " .. currRank )
	    	    		end
	    	    	end	    	    
	    	end
	end
	
	if not self:IsEventScheduled("UpdatingAB") then
		self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1.5, self)
	end
end


function DrDamage:DetermineClass()

	local nomatch
	local deleted

	if englishClass == "WARLOCK" then
		if type( DrDamage.WarlockData ) == "function" then
			spellInfo, talentEffects = DrDamage:WarlockData()
		else
			deleted = true
		end
	elseif englishClass == "MAGE" then
		if type( DrDamage.MageData ) == "function" then
			spellInfo, talentEffects = DrDamage:MageData()
		else
			deleted = true
		end		
	elseif englishClass == "PRIEST" then
		if type( DrDamage.PriestData ) == "function" then
			spellInfo, talentEffects = DrDamage:PriestData()
		else
			deleted = true
		end		
	elseif englishClass == "SHAMAN" then
		if type( DrDamage.ShamanData ) == "function" then
			spellInfo, talentEffects = DrDamage:ShamanData()
		else
			deleted = true
		end	
	elseif englishClass == "DRUID" then
		if type( DrDamage.DruidData ) == "function" then
			spellInfo, talentEffects = DrDamage:DruidData()
		else
			deleted = true
		end
	elseif englishClass == "PALADIN" then
		if type( DrDamage.PaladinData ) == "function" then
			spellInfo, talentEffects = DrDamage:PaladinData()
		else
			deleted = true
		end	
	else
		nomatch = true
	end
	
	if deleted then
		self:Print( "The data file for " .. englishClass:lower() .. "s has been deleted." )
		return false
	elseif nomatch then
		return false
	else
		return true	
	end
end

local function findpattern(text, pattern, start)
	if text and pattern and string.find( text, pattern, start) then
		return string.sub(text, string.find(text, pattern, start))
	else
		return " "
	end
end

function DrDamage:MatchData( data, ... )

	if not data or not ... then
		return false
	end

	if type( data ) == "table" then
		for _, dataName in ipairs( data ) do
			for i = 1, select('#', ...) do
				if dataName == select(i, ...) then
					return true
				end
			end
		end
	else	
		for i = 1, select('#', ...) do
			if data == select(i, ...) then
				return true
			end
		end
	end
	
	return false
end

--Credits to the author of RatingBuster (Whitetooth) for the formula!

function DrDamage:GetRating( hit, convertR, full )
	local playerLevel = UnitLevel("player")
	local base = 14
	local rating = 0

	if hit then
		base = 8
	end

	if playerLevel <= 9 then
		rating = 0
	elseif playerLevel <= 60 then
		rating = base / ( ( playerLevel - 8 ) / 52 )
	elseif playerLevel <= 70 then
		rating = base / ( 131/41 - 3/82 * playerLevel  )
	end

	if full then
		if convertR then
			return ( convertR / rating )
		else
			return rating
		end
	else
		if convertR then
			return DrDamage:Round( convertR / rating, 2 )
		else
			return DrDamage:Round( rating, 2 )
		end
	end
end

function DrDamage:SetSpell( frame, slot )

	if not DrDamage.db.profile.Display then
		do return end
	end
		
	GT:SetSpell( slot, BOOKTYPE_SPELL )
	local spellName, spellRank = GetSpellName( slot, BOOKTYPE_SPELL );
	
	if spellRank and string.len( spellRank ) >= 4 then
		spellRank = tonumber( findpattern( spellRank, "%d+" ))
	end
	
	if not spellRank then
		spellRank = "None"
	end
	
	DrDamage:DoTooltip( frame, spellName, spellRank )

end

function DrDamage:SetAction( frame, slot )

	self.hooks[frame].SetAction(frame, slot)
	
	if not DrDamage.db.profile.Display then
		do return end
	end

	GT:SetAction( slot )	
	local gtype = GetActionInfo( slot )
	local spellName, spellRank = frame:GetSpell()

	if gtype == "spell" or gtype == "macro" then
	
		if spellName then 
		
			if spellRank and string.len( spellRank ) >= 4 then
				spellRank = tonumber( findpattern( spellRank, "%d+" ))
			end
			
			if not spellRank then
				spellRank = "None"
			end
	
			DrDamage:DoTooltip( frame, spellName, spellRank )
		end
	end
end


local function FreeCrits( casts, canCrit, freeCrit, critRate )

	if canCrit and freeCrit and freeCrit > 0 then

		local total = casts

		for i = 1, 5 do	--5 loops enough?
			casts = math_floor( freeCrit * ( critRate / 100 ) * casts )

			if casts == 0 then
				break
			end									

			total = total + casts
		end

		return total
	end

	return casts
end


local TooltipData = {}

function DrDamage:DoTooltip( frame, spellName, spellRank )

	if spellName and spellRank then
  	
		if spellInfo[spellName] then
		
			local baseSpell = spellInfo[spellName][0]
			local tableSpell = spellInfo[spellName][spellRank]
			
			if not tableSpell then
				tableSpell = spellInfo[spellName]["None"]
			end
			
			if tableSpell and not baseSpell.NoTooltip then
				
				local healingSpell, AvgTotal, DPS, DPSC, AvgDmg, MinDmg, MaxDmg, TooltipData = DrDamage:RawNumbers ( tableSpell, spellName, true )
				local DPM, DOOM, SOOM, castsOOM, ManaCost, TrueManaCost, MPS
				local regenCasts = 0
				
				if GT:GetLine( 2 ) and Mana_Cost then
					ManaCost = Deformat( GT:GetLine(2), Mana_Cost )
				end

				if ManaCost then
				
					ManaCost = tonumber( ManaCost )
					
					if ManaCost > 0 then
						
						if TooltipData.FreeCrit and TooltipData.CritRate and TooltipData.CritRate < 100 then
							DPM = DrDamage:Round( ( AvgTotal / ManaCost ) / ( 1 - ( TooltipData.CritRate / 100 ) * TooltipData.FreeCrit ), 1 )
							TrueManaCost = DrDamage:Round( ManaCost * ( 1 - (TooltipData.CritRate / 100 ) * TooltipData.FreeCrit), 1 )
						else
							DPM = DrDamage:Round( AvgTotal / ManaCost, 1 )
						end
					
						local PlayerMana = UnitMana("player")
						local regen = 0
						
						if DrDamage.db.profile.Casts and PlayerMana then
												
							local Cooldown = TooltipData.Cooldown
							local ManaRegen = TooltipData.ManaRegen
							local CastTime = TooltipData.CastTime
							local lowCost

							if Cooldown > 0 and Cooldown <= 10 then
								CastTime = Cooldown
							end

							if Cooldown <= 10 then

								regen = math_floor( PlayerMana / ManaCost ) * CastTime * ManaRegen
								local newCasts = FreeCrits( math_floor( regen / ManaCost ), baseSpell.canCrit, TooltipData.FreeCrit, TooltipData.CritRate )
								regenCasts = regenCasts + newCasts
								
								for i = 1, 5 do
								
									local newRegen = newCasts * CastTime * ManaRegen
									newCasts = FreeCrits( math_floor( newRegen / ManaCost ), baseSpell.canCrit, TooltipData.FreeCrit, TooltipData.CritRate )
									regen = regen + newRegen
									regenCasts = regenCasts + newCasts
									
									if newCasts == 0 then
										break
									end
								end
							end

							castsOOM = FreeCrits( math_floor( PlayerMana / ManaCost ), baseSpell.canCrit, TooltipData.FreeCrit, TooltipData.CritRate )
							SOOM = DrDamage:Round( ( castsOOM + regenCasts ) * CastTime, 1 )
							
							if ( castsOOM + regenCasts ) > 1000 then
								castsOOM = "\226\136\158"
								SOOM = nil
							end
						end
						
						if TrueManaCost then
							MPS = DrDamage:Round( TrueManaCost / TooltipData.CastTime, 1 )
						else
							MPS = DrDamage:Round( ManaCost / TooltipData.CastTime, 1 )
						end
						
						if PlayerMana then
							DOOM = DrDamage:Round( DPM * ( PlayerMana + regen ) )
							
							if DOOM > 1000000 then
								DOOM = "1000000+"
							end
						end
					end
				end
				
				frame:AddLine(" ")
				
				local r, g, b = 1, 0.82745098, 0
				
				if not DrDamage.db.profile.DefaultColor then
					r, g, b = 0, 0.7, 0
				end
				
				local spellType
				local spellAbbr
				
				if healingSpell then
					spellType = L["Heal"]
					spellAbbr = L["H"]
				else
					spellType = L["Dmg"]
					spellAbbr = L["D"]
				end
				
				if DrDamage.db.profile.PlusDmg then
					frame:AddDoubleLine( "+" .. spellType .. L[" (eff.):"], "+" .. DrDamage:Round( TooltipData.SpellDmg * TooltipData.SpellDmgM * TooltipData.DmgM, 1 ), 1, 1, 1, r, g, b  )
				end
				
				if DrDamage.db.profile.Coeffs then
					frame:AddDoubleLine(L["Coeffs:"], TooltipData.DmgM .. " x (" .. TooltipData.SpellDmgM .." x " .. TooltipData.SpellDmg ..")", 1, 1, 1, r, g, b  )
				end
  				
  				if DrDamage.db.profile.DispCrit and TooltipData.CritRate then
  					frame:AddDoubleLine(L["Crit:"], TooltipData.CritRate .. "%", 1, 1, 1, r, g, b )
  				end
  				
  				if DrDamage.db.profile.DispHit and not baseSpell.Unresistable and not healingSpell then
  					frame:AddDoubleLine(L["Hit:"], TooltipData.HitRate .. "%", 1, 1, 1, r, g, b )
  				end  				
  				
  				if not DrDamage.db.profile.DefaultColor then
  					r, g, b = 0.8, 0.8, 0.9
  				end
  				
  				if DrDamage.db.profile.AvgHit then
  					frame:AddDoubleLine(L["Avg:"], AvgDmg .. " (".. MinDmg .."-".. MaxDmg ..")", 1, 1, 1, r, g, b )
  					
  					if baseSpell.Leech and TooltipData.AvgLeech and TooltipData.AvgLeech > AvgDmg then
  						frame:AddDoubleLine(L["Avg Heal:"], TooltipData.AvgLeech, 1, 1, 1, r, g, b )
  					end
  				end
  				
  				if DrDamage.db.profile.AvgCrit and TooltipData.AvgCrit and not baseSpell.sHits then
  					frame:AddDoubleLine(L["Avg Crit:"], TooltipData.AvgCrit .. " (".. TooltipData.MinCrit .."-".. TooltipData.MaxCrit ..")", 1, 1, 1, r, g, b )
  				end
  				
  				if DrDamage.db.profile.Ticks and TooltipData.PerHit then
  					if baseSpell.sHits then
  						frame:AddDoubleLine( L["Hits:"], TooltipData.Hits .. "x ~" .. TooltipData.PerHit, 1, 1, 1, r, g, b )
  						
						if baseSpell.Leech and TooltipData.PerHitHeal and TooltipData.PerHitHeal > TooltipData.PerHit then
							frame:AddDoubleLine( L["Hits Heal:"], TooltipData.Hits .. "x ~" .. TooltipData.PerHitHeal, 1, 1, 1, r, g, b )
						end  						
  					elseif baseSpell.sTicks then
  						frame:AddDoubleLine( L["Ticks:"], TooltipData.Hits .. "x ~" .. TooltipData.PerHit, 1, 1, 1, r, g, b )
  						
						if baseSpell.Leech and TooltipData.PerHitHeal and TooltipData.PerHitHeal > TooltipData.PerHit then
							frame:AddDoubleLine( L["Ticks Heal:"], TooltipData.Hits .. "x ~" .. TooltipData.PerHitHeal, 1, 1, 1, r, g, b )
						end  						
  					end
  				end
  				
  				if DrDamage.db.profile.Extra then 
  				
  					if TooltipData.ChainDmg then
  						frame:AddDoubleLine(L["Avg Chain:"], TooltipData.ChainDmg, 1, 1, 1, r, g, b  )
  					end
  				
  					if TooltipData.IgniteDmg then
  						frame:AddDoubleLine(L["Ignite:"], TooltipData.IgniteDmg, 1, 1, 1, r, g, b )
  					end
  				
   					if TooltipData.DotDmg > 0 then
  						frame:AddDoubleLine(L["DoT:"], TooltipData.DotDmg, 1, 1, 1, r, g, b )
  					end
  				end  				
  				
  				if DrDamage.db.profile.Total and AvgTotal > AvgDmg then
  					frame:AddDoubleLine(L["Avg Total:"], AvgTotal, 1, 1, 1, r, g, b  )
  				end
  				
  				if not DrDamage.db.profile.DefaultColor then
  					r, g, b = 0.3, 0.6, 0.5
  				end
  				
  				  				
  				if DrDamage.db.profile.Next and not baseSpell.NoNext then

  					if TooltipData.NextCrit then
  						frame:AddDoubleLine(L["+1% Crit (+"] .. DrDamage:GetRating() .. "):", "+" .. TooltipData.NextCrit .. " (" .. DrDamage:Round( ( TooltipData.NextCrit * 10 ) / TooltipData.NextTenDmg, 1 ) .. " +" .. spellType .. ")", 1, 1, 1, r, g, b )	
  					end
  					
  					if TooltipData.NextHit then
  						frame:AddDoubleLine(L["+1% Hit (+"] .. DrDamage:GetRating( true ) .. "):", "+" .. TooltipData.NextHit .. " (" .. DrDamage:Round( ( TooltipData.NextHit * 10 ) / TooltipData.NextTenDmg, 1 ) .. " +" .. spellType .. ")", 1, 1, 1, r, g, b )
  					end
  				
  					frame:AddDoubleLine("+10 " .. spellType .. ":", "+" .. TooltipData.NextTenDmg, 1, 1, 1, r, g, b )
  				end
  				
  				if not DrDamage.db.profile.DefaultColor then
  					r, g, b = 0.8, 0.1, 0.1
  				end
  				
  				if DrDamage.db.profile.DPS and not baseSpell.NoDPS then
  				
  					if DPSC ~= DPS then
  						if TooltipData.ExtraDPS then
  							frame:AddDoubleLine( spellAbbr .. L["PS/"] .. spellAbbr .. L["PSC:"], DPS .. "+" .. TooltipData.ExtraDPS .. "/" .. DPSC, 1, 1, 1, r, g, b)
  						else
  							frame:AddDoubleLine( spellAbbr .. L["PS/"] .. spellAbbr .. L["PSC:"], DPS .. "/" .. DPSC, 1, 1, 1, r, g, b)
  						end
      					else
      						if TooltipData.ExtraDPS then
      							frame:AddDoubleLine( spellAbbr .. L["PS:"], DPS .. "+" .. TooltipData.ExtraDPS, 1, 1, 1, r, g, b)
      						else
      							frame:AddDoubleLine( spellAbbr .. L["PS:"], DPS, 1, 1, 1, r, g, b)
      						end
      					end
      					
					if TooltipData.DPSCD then
						frame:AddDoubleLine( spellAbbr .. L["PS (CD):"], TooltipData.DPSCD, 1, 1, 1, r, g, b)
					end  					
      				end
      				
      				if DrDamage.db.profile.DPM and DPM and not baseSpell.NoDPM then
      					frame:AddDoubleLine( spellAbbr .. L["PM:"], DPM, 1, 1, 1, r, g, b )
      				end
      				
      				if DrDamage.db.profile.Doom and DOOM and not baseSpell.NoDoom then
      					frame:AddDoubleLine( spellType .. L[" until OOM:"], DOOM, 1, 1, 1, r, g, b )
      				end
      				
      				if DrDamage.db.profile.Casts and castsOOM then
					if TooltipData.Cooldown <= 10 and not baseSpell.eDot then
						if SOOM then
							if regenCasts > 0 then
								frame:AddDoubleLine(L["Casts (rgn):"], castsOOM .. "+" .. regenCasts .." (" .. SOOM .. "s)", 1, 1, 1, r, g, b )
							else
								frame:AddDoubleLine(L["Casts (rgn):"], castsOOM .." (" .. SOOM .. "s)", 1, 1, 1, r, g, b )
							end
						else
							frame:AddDoubleLine(L["Casts (rgn):"], castsOOM, 1, 1, 1, r, g, b )
						end
					else
						frame:AddDoubleLine(L["Casts:"], castsOOM, 1, 1, 1, r, g, b )
					end
      				end
      				
      				if DrDamage.db.profile.ManaUsage then
      					if TrueManaCost then
      						frame:AddDoubleLine( L["True Mana Cost:"], TrueManaCost, 1, 1, 1, r, g, b )
      					end
      					
      					if MPS then
      						frame:AddDoubleLine( L["MPS:"], MPS, 1, 1, 1, r, g, b )
      					end
      				end
      				
				DrDamage:ClearTable( TooltipData )
      				
      				frame:Show()			
			end  			
		end	  
	end
end
	
local ABtable, globalbutton
local BT3, Bongos, Trinity, FB2, IB, Cogs, DAB, NAB, CTBar, blizzAB, noCustomAB
local defaultBars = { "ActionButton", "BonusActionButton", "MultiBarBottomLeftButton", "MultiBarBottomRightButton","MultiBarRightButton", "MultiBarLeftButton", }
		

function DrDamage:OnAttributeChanged( object, name, value )

	if name == "state-alt" and DrDamage.db.profile.UpdateAlt  or name == "state-ctrl" and DrDamage.db.profile.UpdateCtrl or name == "state-shift" and DrDamage.db.profile.UpdateShift then
			if not self:IsEventScheduled("UpdatingAB") then
				self:ScheduleEvent("UpdatingAB", function() DrDamage:UpdateAB( false, true ) end, 0.5, self)
				visualChange = true
			end
	end
		
    	--return self.hooks[object].OnAttributeChanged( object, name, value )
end

function DrDamage:DetermineAB()

	if IsAddOnLoaded("Bartender3") then
		BT3 = true
		globalbutton = "BT3Button"
		
		if DrDamage.db.profile.UpdateAlt or DrDamage.db.profile.UpdateCtrl or DrDamage.db.profile.UpdateShift  then
		
			DrDamage:SecureHook( "SecureStateHeader_OnAttributeChanged", "OnAttributeChanged" )
		
			--[[
			if BT3Bar1 then
				DrDamage:HookScript( BT3Bar1, "OnAttributeChanged" )
			end
			if BT3Bar2 then
				DrDamage:HookScript( BT3Bar2, "OnAttributeChanged" )
			end
			if BT3Bar3 then
				DrDamage:HookScript( BT3Bar3, "OnAttributeChanged" )
			end
			if BT3Bar4 then
				DrDamage:HookScript( BT3Bar4, "OnAttributeChanged" )
			end
			if BT3Bar5 then
				DrDamage:HookScript( BT3Bar5, "OnAttributeChanged" )
			end
			if BT3Bar6 then
				DrDamage:HookScript( BT3Bar6, "OnAttributeChanged" )
			end
			if BT3Bar7 then
				DrDamage:HookScript( BT3Bar7, "OnAttributeChanged" )
			end
			if BT3Bar8 then
				DrDamage:HookScript( BT3Bar8, "OnAttributeChanged" )
			end
			if BT3Bar9 then
				DrDamage:HookScript( BT3Bar9, "OnAttributeChanged" )
			end
			if BT3Bar10 then
				DrDamage:HookScript( BT3Bar10, "OnAttributeChanged" )
			end
			--]]
		end
		
		if not ABtable then
			ABtable = Bartender3.actionbuttons
		end
	elseif IsAddOnLoaded( "Bongos" ) then
		Bongos = true
		globalbutton = "BActionButton"
	elseif IsAddOnLoaded("TrinityBars") then
		Trinity = true
		globalbutton = "TrinityActionButton"
	elseif IsAddOnLoaded( "FlexBar2" ) then
		FB2 = true
	elseif IsAddOnLoaded( "CogsBar" ) then
		Cogs = true
		globalbutton = "CogsBarButton"
	elseif IsAddOnLoaded( "InfiniBar" ) then
		IB = true		
	elseif IsAddOnLoaded("DiscordActionBars") then
		DAB = true
		globalbutton = "DAB_ActionButton_"
	elseif IsAddOnLoaded("Nurfed") then
		NAB = true
		globalbutton = "Nurfed_Button"
	elseif IsAddOnLoaded("CT_BarMod") then
		CTBar = true
		if not ABtable then
			ABtable = CT_BarMod.actionButtonList
		end
	else
		blizzAB = true
		noCustomAB = true
	end
end
	
function DrDamage:UpdateAB( disable, hide, manaUpdate, updateSpell )

	--[[	Used for debugging updates.
	
	if updateSpell then
		self:Print( "Update AB: Spell" )
		
		if type( updateSpell ) == "table" then
			for _, gar in ipairs( updateSpell ) do
				self:Print( "Updating: " .. gar )
			end
		else
			self:Print( "Updating: " .. updateSpell )
		end
	else
		self:Print( "Update AB: All" )
	end
	--]]

	DrDamage:TriggerEvent("DrDamage_Update", updateSpell )
	
	if DrDamage.db.profile.BlizzardAB and not blizzAB then
		blizzAB = true
	end

	if ( FB2 or IB ) and not blizzAB then
		do return end
	end
	
	if CTBar then
		for i = 1, 12 do
			DrDamage:ProcessButton( _G["ActionButton"..i], i, true, disable, hide, manaUpdate, updateSpell )
		end		
		
		for key, value in pairs( ABtable ) do 
			DrDamage:ProcessButton( value.button, key, nil, disable, hide, manaUpdate, updateSpell )
		end
	elseif globalbutton then
		for i = 1, 120 do	
			DrDamage:ProcessButton( _G[globalbutton..i], i, nil, disable, hide, manaUpdate, updateSpell )
		end
	end
	
	if blizzAB then
	
		local start = 1
		
		if CTBar then
			start = 2
		end
		
		if not DrDamage.db.profile.BlizzardAB and not noCustomAB then
			disable = true
			blizzAB = false
		end			
	
		for bar = start, 6 do
			for number = 1, 12 do
				DrDamage:ProcessButton( _G[defaultBars[bar]..number], nil, nil, disable, hide, manaUpdate, updateSpell, true )
			end
		end		
	end
					
	updateABSpell = nil
	updatingSpell = nil
	visualChange = nil
end

local function HideText(...)
	for i = 1, select('#', ...) do
		local frame = select(i, ...)
		if frame and frame:GetName() == "DrDamageText" and frame.text then
			frame.text:Hide()
			return frame
		end
	end
end

function DrDamage:ProcessButton( button, index, main, disable, hide, manaUpdate, updateSpell, defaultBar )

	if not button then
		do return end
	end

	local textFrame

	if disable or hide then
		textFrame = HideText(button:GetChildren())
	end
	
	if DrDamage.db.profile.DisplayType == "Off" or disable then
		do return end
	end

	if button:IsVisible() then

		local id, spellName, spellRank

		if BT3 and ABtable[index] and ABtable[index].state == "used" then
			id = ABtable[index].object:PagedID()
		elseif Bongos then
			id = button:GetPagedID()
		elseif Trinity then
			id = button:GetID()
		elseif Cogs then
			id = SecureButton_GetModifiedAttribute( button, "action", SecureStateChild_GetEffectiveButton( button ) )
		elseif DAB then
			id = button:GetActionID()
		elseif NAB then
			if button.spell then
				local pos = string.find( button.spell, "%(" )

				if pos then
					pos = pos - 1
				else
					pos = string.len( button.spell )
				end

				spellName = string.sub( button.spell, 1, pos )
				spellRank = findpattern( button.spell, "%d+" )
			end
		elseif CTBar then
			if main then
				id = ActionButton_GetPagedID( button )
			elseif ABtable[index] and ABtable[index].hasAction then
				id = index
			end
		end
		
		if defaultBar then
			id = ActionButton_GetPagedID( button )
		end

		if id then
			DrDamage:CheckAction( id, button, manaUpdate, updateSpell, nil, nil, textFrame )
		elseif spellName and spellRank then
			DrDamage:CheckAction( nil, button, manaUpdate, updateSpell, spellName, spellRank, textFrame )
		end
	end
end


local displayTypeTable = { ["Avg"] = 2, ["DPS"] = 3, ["DPSC"] = 4, ["AvgHit"] = 5, ["Min"] = 6, ["Max"] = 7, }

function DrDamage:CheckAction( id, button, manaUpdate, updateSpell, spellName, spellRank, textFrame )

	local gtype, pid

	if id then
	
		gtype, pid = GetActionInfo(id)

		if gtype == "spell" or gtype =="macro" then

			if gtype == "spell" then
				spellName, spellRank = GetSpellName( pid, BOOKTYPE_SPELL );
			else
				GT:SetAction( id )
				spellName = GT:GetLine(1, false)
				spellRank = GT:GetLine(1, true)

				if not spellName and not spellRank and button then
					local macroText = getglobal(button:GetName() .. "Name")

					if not macroText:IsVisible() then
						macroText:Show()
					end			
				end
			end
		end
	end
			
	if spellName then
	
		if spellName == BS["Shoot"] then
			return DrDamage:WandSlot( button, textFrame )
		end
	
		if spellRank and spellInfo[spellName] then

			if updateSpell then

				if not DrDamage:MatchData( updateSpell, spellName ) then
					do return end
				end

				--self:Print( "Spell matched." )
				updateABSpell = false
			end

			local baseSpell = spellInfo[spellName][0]
			local tableSpell = spellInfo[spellName][ tonumber( findpattern( spellRank, "%d+" ) ) ]

			if not tableSpell then
				tableSpell = spellInfo[spellName]["None"]
			end

			if tableSpell then

				local healingSpell

				if baseSpell.Healing then
					healingSpell = true
				elseif DrDamage:MatchData( baseSpell.School, "Healing" ) then
					healingSpell = true
				end

				if gtype == "macro" and button then

					local macroText = getglobal(button:GetName() .. "Name")

					if macroText then
						macroText:Hide()
					end

					if IsAddOnLoaded("CT_BarMod") then
						if button.name then
							button.name:SetText( "" )
						end
					end
				end

				local didCasts	

				if id and ( DrDamage.db.profile.CastsLeft and healingSpell or DrDamage.db.profile.CastsLeftDmg and not healingSpell ) then

					GT:SetAction( id )
					local ManaCost

					if GT:GetLine( 2 ) and Mana_Cost then
						ManaCost = Deformat( GT:GetLine(2), Mana_Cost )
					end				

					if ManaCost then

						local PlayerMana = UnitMana("player")				
						ManaCost = tonumber( ManaCost )
						didCasts = true

						local text

						if ManaCost > 0 then
							text = math_floor( PlayerMana / ManaCost )
						else
							text = "Inf."
						end

						if button then
							DrDamage:CreateABtext( button, text, healingSpell, textFrame )
						else
							return DrDamage:Round( text )
						end
					end
				end

				if manaUpdate then
					do return end
				end

				if not didCasts then

					local displayType = DrDamage.db.profile.DisplayType
					local selector = displayTypeTable[displayType] or 2

					if baseSpell.NoDPS and selector ~= 5 then
						selector = 2
					end

					local text

					if displayType == "Off" then
						text = ""
					else
						text = select( selector, DrDamage:RawNumbers ( tableSpell, spellName ) )
					end

					if button and text then
						DrDamage:CreateABtext( button, text, healingSpell, textFrame )
					elseif text then
						return DrDamage:Round( text )
					end
				end
			end				
		end
	end
end

local function FindText(...)
	
	for i = 1, select('#', ...) do
		local frame = select(i, ...)
		
		if frame and frame:GetName() == "DrDamageText" and frame.text then
    			return frame
    		end	
    	end
end

function DrDamage:CreateABtext( button, cDmg, HealingSpell, frame )

	local positionMod = 0
	cDmg = DrDamage:Round( cDmg )
	
	if cDmg and string.len( cDmg ) then
		positionMod = math_floor( (4 - string.len( cDmg ) ) * 3.5 )
	end

	if not frame then
		frame = FindText( button:GetChildren() )
	end
	
	if frame then
	
		if visualChange then
			frame.text:SetFont( STANDARD_TEXT_FONT, DrDamage.db.profile.FontSize, "OUTLINE" )

			if HealingSpell then
				frame.text:SetTextColor( DrDamage.db.profile.FontColorHeal.r, DrDamage.db.profile.FontColorHeal.g, DrDamage.db.profile.FontColorHeal.b )
			else
				frame.text:SetTextColor( DrDamage.db.profile.FontColorDmg.r, DrDamage.db.profile.FontColorDmg.g, DrDamage.db.profile.FontColorDmg.b )
			end
		end

		frame.text:SetPoint("BOTTOMLEFT", frame, "BOTTOM", (DrDamage.db.profile.FontXPosition - 15) + positionMod, (DrDamage.db.profile.FontYPosition + 5) )
		frame.text:SetText( cDmg )
		frame.text:Show()
	else
		local dpsText = CreateFrame("Frame", "DrDamageText", button )
		dpsText:SetAllPoints( button )
		dpsText:SetFrameLevel(dpsText:GetFrameLevel() + 1)
		dpsText.text = dpsText:CreateFontString( nil, "OVERLAY" )
		dpsText.text:SetFontObject(GameFontNormalLarge)
		dpsText.text:SetPoint("BOTTOMLEFT", dpsText, "BOTTOM", (DrDamage.db.profile.FontXPosition - 15) + positionMod, (DrDamage.db.profile.FontYPosition + 5) )
		dpsText.text:SetFont(STANDARD_TEXT_FONT, DrDamage.db.profile.FontSize, "OUTLINE")
		
		if HealingSpell then
			dpsText.text:SetTextColor( DrDamage.db.profile.FontColorHeal.r, DrDamage.db.profile.FontColorHeal.g, DrDamage.db.profile.FontColorHeal.b )
		else
			dpsText.text:SetTextColor( DrDamage.db.profile.FontColorDmg.r, DrDamage.db.profile.FontColorDmg.g, DrDamage.db.profile.FontColorDmg.b )
		end
		
		dpsText.text:SetText( cDmg )
		dpsText.text:Show()
		
		--Fix to properly inherit parent alpha
		if button:GetAlpha() == 0 then
			button:SetAlpha( 0.01 )
			button:SetAlpha( 0 )
		end
		
	end
end

function DrDamage:WandSlot( button, textFrame )

	local selector = displayTypeTable[DrDamage.db.profile.DisplayType] or 2
	local text

	if displayType == "Off" then
		text = ""
	else
		text = select( selector, DrDamage:WandDamage() )
	end
	
	if button and text then
		DrDamage:CreateABtext( button, text, nil, textFrame )
	elseif text then
		return text
	end
end

function DrDamage:WandDamage()
	local speed, lowDmg, hiDmg = UnitRangedDamage( "player" )
	local avgDmg = (lowDmg + hiDmg) / 2
	local avgTotal = avgDmg * ( 1 + ( GetRangedCritChance() / 100) * 0.5 )
	local DPS = DrDamage:Round( avgTotal / speed )
	
	return nil, DrDamage:Round( avgTotal ), DPS, DPS,  DrDamage:Round( avgDmg ), DrDamage:Round( lowDmg ), DrDamage:Round( hiDmg )
end

local hitDataMOB = { 95, 94, 83, 72, 61, 50, 39, 28, 17, 6 }
local hitDataPlayer = { 95, 94, 87, 80, 73, 66, 59, 52, 45, 38 }
local lastTargetLevel
local lastSpellHit

function DrDamage:SpellHit()

	local targetLevel = UnitLevel( "target" )
	local playerLevel = UnitLevel( "player" )
	local hitPerc = 99
	
	if DrDamage.db.profile.HitTarget then
		
		local levelDiff = 0

		if playerLevel >= 60 and targetLevel == -1 then
			levelDiff = 3
		elseif targetLevel == -1 then
			levelDiff = 10
		elseif targetLevel > playerLevel then
			levelDiff = targetLevel - playerLevel
		elseif targetLevel == playerLevel then
			hitPerc = 96
		end
		
		if levelDiff > 0 and levelDiff <= 10 then
		
			if UnitIsPlayer( "target" ) then
				hitPerc = hitDataPlayer[levelDiff]
			else
				hitPerc = hitDataMOB[levelDiff]
			end
		end
		
		lastTargetLevel = targetLevel
		lastSpellHit = hitPerc
		
	elseif not DrDamage.db.profile.HitTarget then
		
		if DrDamage.db.profile.TargetLevel >= 1 then
			
			if DrDamage.db.profile.TargetPlayer then
					hitPerc = hitDataPlayer[ math_floor( DrDamage.db.profile.TargetLevel ) ]
			else
					hitPerc = hitDataMOB[ math_floor( DrDamage.db.profile.TargetLevel ) ]
			end
					
		elseif math_floor( DrDamage.db.profile.TargetLevel ) == 0 then
			hitPerc = 96
		end
		
		lastTargetLevel = DrDamage.db.profile.TargetLevel
		lastSpellHit = hitPerc		
	end
	
	if select( 2, UnitRace( "player" ) ) == "Draenei" and ( englishClass == "MAGE" or englishClass == "PRIEST" or englishClass == "SHAMAN" ) then
		hitPerc = hitPerc + 1
	end
		
	return hitPerc
end

function DrDamage:GetPlusDamage( school )

	if school == "Holy" then
  		return GetSpellBonusDamage(2)
	elseif school == "Fire" then
  		return GetSpellBonusDamage(3)
  	elseif school == "Nature" then
  		return GetSpellBonusDamage(4)
  	elseif school == "Frost" then
  		return GetSpellBonusDamage(5)
  	elseif school == "Shadow" then
  		return GetSpellBonusDamage(6)  		
   	elseif school == "Arcane" then
  		return GetSpellBonusDamage(7)
  	else
  		return GetSpellBonusDamage(1)
  	end	
end

function DrDamage:SpellCrit( school )
	
	if school == "Holy" then
  		return GetSpellCritChance(2)
	elseif school == "Fire" then
  		return GetSpellCritChance(3)
  	elseif school == "Nature" then
  		return GetSpellCritChance(4)
  	elseif school == "Frost" then
  		return GetSpellCritChance(5)
  	elseif school == "Shadow" then
  		return GetSpellCritChance(6)  		
   	elseif school == "Arcane" then
  		return GetSpellCritChance(7)
  	else
  		return GetSpellCritChance(1)
  	end
end

function DrDamage:TotalAP()
	local baseAP, posBuff, negBuff = UnitAttackPower( "player" )
	return baseAP + posBuff + negBuff
end

function DrDamage:BaseMelee()
	local rangeMin, rangeMax, _, _, _, _, percMod = UnitDamage( "player" )
	local atkPower = DrDamage:TotalAP()
	local baseMin = rangeMin / percMod - ( atkPower / 14 ) * UnitAttackSpeed( "player" )
	local baseMax = rangeMax / percMod - ( atkPower / 14 ) * UnitAttackSpeed( "player" )
	
	return baseMin, baseMax
end

function DrDamage:GetSpellInfo( spellName, spellRank )

	if spellName and spellRank then
	
		if spellInfo[spellName] then
		
			local tableSpell = spellInfo[spellName][spellRank]

			if not tableSpell then
				tableSpell = spellInfo[spellName]["None"]
			end

			if tableSpell then
				return tableSpell
			end
		end
	end
end

local CalculationResults = {}

function DrDamage:RawNumbers( spell, lSpellName, tooltip )

	if not spell or not lSpellName then
		do return end
	end
	
	local spellName = lSpellName
	local baseSpell = spellInfo[spellName][0]
	
	if BS:HasReverseTranslation( spellName ) then
		spellName = BS:GetReverseTranslation( spellName )
	end		

	local healingSpell = false
	local spellSchool 
	local spellType
	
	if type( baseSpell.School ) == "table" then
		spellSchool = baseSpell.School[1]
		spellType = baseSpell.School[2]
		
		if spellType == "Healing" then
			healingSpell = true
		end
	else
		spellSchool = baseSpell.School
	end
	
	if baseSpell.Healing then
		healingSpell = true
	end	
	
	local critPerc = DrDamage:SpellCrit( spellSchool ) 		--Calculate crit
	local hitPerc = lastSpellHit
	local _, manaRegenCasting = GetManaRegen( "player" )
	
	--Calculate base hit chance only if needed:
	local targetLevel = UnitLevel( "target" )
	
	if DrDamage.db.profile.HitTarget and targetLevel ~= lastTargetLevel then
		hitPerc = DrDamage:SpellHit() 		
	elseif not DrDamage.db.profile.HitTarget and DrDamage.db.profile.TargetLevel ~= lastTargetLevel then
		hitPerc = DrDamage:SpellHit()
	end
		
	hitPerc = hitPerc + GetCombatRatingBonus( 8 )   --Add Gear
	
	local spellDmg
	
	if healingSpell then
		spellDmg = GetSpellBonusHealing() 			-- +healing
	else
		spellDmg = DrDamage:GetPlusDamage( spellSchool ) 	--+all spelldamage
	end
	
	local critM = 0.5			--Crit multiplier
	local bNukeDmg = 0			--Modifier for spells with a DOT portion
	local dmgM = 1				--Damage multiplier (total)
	local bDmgM = 1				--Base damage Modifier
	local spellDmgM				--Spell damage multiplier
	local dotSpellDmgM = 0			--Dot portion spelldamage modifier
	local igniteM = 0			--Ignites % of total
	local cooldown = 0			--Spell's cooldown
	local finalMod = 0			--Modifier to final damage +/-
	local dotFinalMod = 0			--Modifier to final dot damage +/-
	local LeechBonus = 1			--Variable for spells with healing in addition to damage
	local minDam = spell[1]			--Spell initial base min damage
	local maxDam = spell[2]			--Spell initial base max damage
	local chainFactor = 0			--Chain effect spells
	local BuffTalentRank = 0		--Special variable for buffs with no API rank (MD for example)
	local BuffTalentRank2 = 0		--As above
	local BuffTalentRank3 = 0		--As above
	local impSanctity			--Improved Sanctity Aura
	local freeCrit				--Illumination / Shammy clearcasting
	local jotcBonus				--Judgement of the Crusader debuff
	local SSRank				--Special variable for Soul Siphon
	local ShatterRank			--Special variable for Shatter
	
	
	--Physical type spells
	if DrDamage:MatchData( baseSpell.School, "Physical" ) then
		critPerc = GetCritChance()
		critM = 1
	end
	
	--Random item bonuses to base values not supplied by API
	if DrDamage.RelicSlot then
		if DrDamage.RelicSlot[spellName] then
			local data = DrDamage.RelicSlot[spellName]
			local count = #data
			
			if count then
				for i = 1, count - 1, 2 do
					if data[i] and data[i+1] then
						if IsEquippedItem( data[i] ) then
						
							local modType = data["ModType"..((i+1)/2)]
							
							if not modType then
								spellDmg = spellDmg + data[i+1]
							elseif modType == "Base" then
								minDam = minDam + data[i+1]
								maxDam = maxDam + data[i+1]
							elseif modType == "Final" then
								finalMod = finalMod + data[i+1]
							elseif modType == "Dot" then
								dotFinalMod = dotFinalMod + data[i+1]
							end
						end
					end
				end
			end
		end
	end
	
	if not healingSpell then
		if UnitCreatureType( "target" ) == "Undead" then
			if IsEquippedItem( 19182 ) then								--Rune of the Dawn
				if not baseSpell.NoBuffs then
					spellDmg = spellDmg + 48
				end	
			end
		elseif UnitCreatureType( "target" ) == "Demon" then
			if IsEquippedItem( 30787 ) then								--Illidari-Bane Mageblade
				if not baseSpell.NoBuffs then
					spellDmg = spellDmg + 185
				end
			end		
		end	
	end	
	
	--Manual variables from profile:
	if DrDamage.db.profile.ManualDmg then
		
		if not DrDamage.db.profile.ManualPlus then
			
			if healingSpell then
				spellDmg = DrDamage.db.profile.Healing
			else
				spellDmg = DrDamage.db.profile.SpellDamage
			end
			
			critPerc = critPerc - GetCombatRatingBonus(11) + DrDamage:GetRating( false, DrDamage.db.profile.CritRating, true )
			hitPerc = hitPerc - GetCombatRatingBonus(8) + DrDamage:GetRating( true, DrDamage.db.profile.HitRating, true )
		else
			if healingSpell then
				spellDmg = spellDmg + DrDamage.db.profile.Healing
			else
				spellDmg = spellDmg + DrDamage.db.profile.SpellDamage
			end
			
			critPerc = critPerc + DrDamage:GetRating( false, DrDamage.db.profile.CritRating, true )
			hitPerc = hitPerc + DrDamage:GetRating( true, DrDamage.db.profile.HitRating, true )
		end
		
		if not healingSpell then
		
			if ( spellSchool == "Shadow" ) then
				spellDmg = spellDmg + DrDamage.db.profile.ShadowDamage --add shadowdamage
			elseif ( spellSchool == "Fire" ) then
				spellDmg = spellDmg + DrDamage.db.profile.FireDamage --add firedamage
			elseif ( spellSchool == "Nature" ) then
				spellDmg = spellDmg + DrDamage.db.profile.NatureDamage --add naturedamage
			elseif ( spellSchool == "Frost" ) then
				spellDmg = spellDmg + DrDamage.db.profile.FrostDamage --add frostdamage
			elseif ( spellSchool == "Arcane" ) then
				spellDmg = spellDmg + DrDamage.db.profile.ArcaneDamage --add arcanedamage
			elseif ( spellSchool == "Holy" ) then
				spellDmg = spellDmg + DrDamage.db.profile.HolyDamage --add holydamage			
			end
		else
			spellDmg = spellDmg + DrDamage.db.profile.Healing
		end
		
		if spellDmg < 0 then
			spellDmg = 0
		end

		if critPerc < 0 then
			critPerc = 0
		end

		if hitPerc < 0 then
			hitPerc = 0
		end
	end
							
	if baseSpell.chainFactor then
		chainFactor = baseSpell.chainFactor
	end
	
	if baseSpell.Cooldown then
		cooldown = baseSpell.Cooldown
	end

	local castTime = 1.5
	
	if spell.castTime then
		castTime = spell.castTime
	elseif baseSpell.castTime then
		castTime = baseSpell.castTime
	end
	
	local eDuration = 1
	
	if spell.eDuration then
		eDuration = spell.eDuration
	elseif baseSpell.eDuration then
		eDuration = baseSpell.eDuration
	end
	
	if baseSpell.bonusFactor then
		spellDmgM = baseSpell.bonusFactor
	elseif baseSpell.eDot then 
		spellDmgM = ( eDuration / 15)
	else
		spellDmgM = ( castTime / 3.5) 
	end
	
	local sHits
	
	if spell.sHits then
		sHits = spell.sHits
	elseif baseSpell.sHits then
		sHits = baseSpell.sHits
	end
	
	--Temporary talent calculation variables
	local dmgMAdd = 0			--Damage multiplier additive portion
	local spellDmgMAdd = 0			--Spell damage multiplier additive portion
				
	for talentName, talentRank in pairs( talents ) do
		
		if talentName then
						
			if talentEffects[talentName] then
				
				local effectTalent = talentEffects[talentName]
				
				local function DoCalc( ModType, talentSpell, talentValue, add, multiply )
				
					if ( talentSpell == "All" or talentSpell == spellSchool ) and not baseSpell.NoSchoolTalents or talentSpell == spellType or talentSpell == spellName then
					
						if not ModType then
						
							if add then
								dmgMAdd = dmgMAdd + talentValue
							else
								dmgM = dmgM * ( 1 + talentValue )	
							end					
						elseif ModType == "SpellDamage" then
						
							if multiply then
								spellDmgM = spellDmgM * ( 1 + talentValue )
							else
								spellDmgMAdd = spellDmgMAdd + talentValue
							end
						elseif ModType == "Crit" then
							critPerc = critPerc + talentValue
						elseif ModType == "Hit" then
							hitPerc = hitPerc + talentValue							
						elseif ModType == "CastTime" then
							
							if baseSpell.CastMod then
								castTime = castTime - baseSpell.CastMod * talentValue
							end
							
							if baseSpell.CastModDmg then
								spellDmgM = spellDmgM * ( 1 - baseSpell.CastModDmg * talentValue )
							end
						elseif ModType == "BaseNuke" then
							bNukeDmg = bNukeDmg + talentValue
						elseif ModType == "BaseDamage" then
							bDmgM = bDmgM + talentValue
						elseif ModType == "DotSpellDamage" then
							dotSpellDmgM = dotSpellDmgM + talentValue
						elseif ModType == "CritMultiplier" then
							critM = critM + talentValue
						elseif ModType == "SpellDuration" then
							eDuration = eDuration + talentValue
						elseif ModType == "Cooldown" then
							cooldown = cooldown - talentValue
						elseif ModType == "Ignite" then
							igniteM = talentValue	
						elseif ModType == "Shatter" then
							ShatterRank = talentValue							
						elseif ModType == "BuffTalentRank" then
							BuffTalentRank = talentValue
						elseif ModType == "BuffTalentRank2" then
							BuffTalentRank2 = talentValue
						elseif ModType == "BuffTalentRank3" then
							BuffTalentRank3 = talentValue							
						elseif ModType == "FreeCrit" then
							freeCrit = talentValue
						elseif ModType == "Soul Siphon" then
							SSRank = talentValue
						elseif ModType == "Crusade" then
							if not healingSpell then
								if UnitCreatureType( "target" ) == "Undead"
								or UnitCreatureType( "target" ) == "Demon"
								or UnitCreatureType( "target" ) == "Humanoid"
								or UnitCreatureType( "target" ) == "Elemental" then
									dmgM = dmgM * ( 1 + talentValue )
								end
							end
						elseif ModType == "ImpSanctity" then
							impSanctity = talentValue							
						elseif ModType == "Hits" then
							if sHits then
								sHits = sHits + talentValue
							end
						elseif ModType == "OneHanded" then
							if IsEquippedItemType( "One-Handed Axes" ) or IsEquippedItemType( "One-Handed Maces" ) or IsEquippedItemType( "One-Handed Swords"  ) then
								dmgM = dmgM * ( 1 + talentValue )
							end
						end
					end				
				
				end
						 
				local talentValue
				
				if type( effectTalent.Effect ) == "table" then
					talentValue = effectTalent.Effect[talentRank]
				else
					talentValue = talentRank * effectTalent.Effect
				end
				
				for j, talentSpell in ipairs( effectTalent.Spells ) do
					DoCalc( effectTalent.ModType, talentSpell, talentValue, effectTalent.Add, effectTalent.Multiply  )
				end
									
				if effectTalent.Spells2 then
				 
					if effectTalent.Effect2 then
						talentValue = talentRank * effectTalent.Effect2
					end
					
					for j, talentSpell in ipairs( effectTalent.Spells2 ) do
						DoCalc( effectTalent.ModType2, talentSpell, talentValue, effectTalent.Add2, effectTalent.Multiply2 )
					end
				end	
			end
		end		
	end
	
	--Add up additive components:
	dmgM = dmgM + dmgMAdd
	spellDmgM = spellDmgM + spellDmgMAdd
	
	
	local bolBonus			--Blessing of Light active on target
	
	local function BuffCalc( bName, bData, bApps, bTexture, bRank, bIndex )
	
		local modType = bData.ModType
		
		if bData.Ranks then
			local buffRank

			if bRank and string.len( bRank ) >= 6 then
				buffRank = tonumber( findpattern( bRank, "%d+" ) )
			end

			if not buffRank then
				buffRank = bData.Ranks
			end

			if not modType then
				dmgM = dmgM * ( 1 + buffRank * bData.Value )				
			elseif modType == "RankTable" then
				spellDmg = spellDmg + bData.Value[buffRank]
			elseif modType == "RankTable2" then
				if spellName == bData.Spell[1] then
					if bData.BoL then
						finalMod = finalMod + bData.Value[buffRank]
						bolBonus = true	
					else
						spellDmg = spellDmg + bData.Value[buffRank]
					end
				elseif spellName == bData.Spell[2] then
					if bData.BoL then
						finalMod = finalMod + bData.Value2[buffRank]
						bolBonus = true
					else
						spellDmg = spellDmg + bData.Value2[buffRank]
					end
				end
			elseif modType == "JotC" then
				jotcBonus = bData.Value[buffRank]											
			end
		elseif bData.Apps then
			if bApps and bApps > 0 then
				if modType == "Crit" then
					critPerc = critPerc + bApps * bData.Value
				elseif modType == "BaseDamage" then
					bDmgM = bDmgM + bApps * bData.Value
				elseif modType == "CastTime" then
					if bData.Value < 0 then
						castTime = castTime + bApps * bData.Value
					end
				elseif not modType then
					if GetRealZoneText() == "Naxxramas" and ( bName == "Positive Charge" or bName == "Negative Charge" ) then
						dmgM = dmgM * ( 1 + bApps * ( bData.Value / 10 ) )
					else
						dmgM = dmgM * ( 1 + bApps * bData.Value )
					end			
				end
			end
		elseif bData.Texture and bTexture and string.find( bTexture, bData.Texture ) then
			dmgM = dmgM * ( 1 + bData.Value )
		elseif bData.Apps2 then
			dmgM = dmgM * ( 1 + bApps * bData.Value2 )
		elseif modType == "CastTime" then
			if bData.Value < 0 then
				castTime = castTime + bData.Value
			else
				castTime = castTime * bData.Value
			end
		elseif modType == "BaseDamage" then
			bDmgM = bDmgM + bData.Value
		elseif modType == "SpellDamage" then
			spellDmgM = spellDmgM + bData.Value
		elseif modType == "Crit" then
			critPerc = critPerc + bData.Value
		elseif modType == "Hit" then
			hitPerc = hitPerc + bData.Value
		elseif not modType then
			dmgM = dmgM * ( 1 + bData.Value )
		end	
	end
	
	--Start calculating Buffs/Debuffs that affects damage not provided by blizzard API.
	local afflCount				--Count (affliction) effects for life drain	
	
	--DEBUFF -- DAMAGE/HEALING -- TARGET
	if DrDamage.Debuffs and not baseSpell.NoDebuffs then
	
		local socStun			--Target is stunned?
		local incinerateBonus 		--Immolate bonus to incinerate only once
		local iceLanceBonus		--Ice Lance bonus only once
		local shatterBonus		--Shatter bonus only once
		local debuffBonus		--Bonus from certain debuffs (eg. druid 4 part set bonus)
		
		local debuffTarget
		
		if not healingSpell then
			debuffTarget = "target"
		elseif UnitIsFriend( "player", "target" ) then
			debuffTarget = "target"
		elseif UnitIsFriend( "player", "targettarget" ) then
			debuffTarget = "targettarget" 
		end		

		if debuffTarget then

			for buffName, apps, _, texture, rank in SEA:DebuffIter( debuffTarget ) do

				if BS:HasReverseTranslation( buffName ) then
					buffName = BS:GetReverseTranslation( buffName )
				end

				if DrDamage.Debuffs[buffName] then

					local buffData = DrDamage.Debuffs[buffName]
					local modType = buffData.ModType				

					if SSRank then
						if buffData.Affliction then
							if not afflCount then
								afflCount = 1
							else
								afflCount = afflCount + 1
							end
						end
					end

					if spellName == "Seal of Command" and not socStun and buffData.Stun then
						socStun = true
						bDmgM = bDmgM + 1
					end

					if not healingSpell 
						and ( DrDamage:MatchData( buffData.School, spellSchool ) and not baseSpell.NoSchoolBuffs
						or not buffData.School and not buffData.Spell )
						or DrDamage:MatchData( buffData.School, spellType )
					or healingSpell and buffData.School == "Healing"
					or DrDamage:MatchData( buffData.Spell, lSpellName )
						and not buffData.Stun
						and not buffData.Affliction
					then
						if modType == "Special" then
							if buffName == "Immolate" and not incinerateBonus then
								incinerateBonus = true
								minDam = minDam + spell.incinerateMin
								maxDam = maxDam + spell.incinerateMax
							elseif ( buffName == "Frost Nova" or buffName == "Frostbite" ) then

								if ShatterRank and ShatterRank > 0 and not shatterBonus then
									shatterBonus = true
									critPerc = critPerc + 10 * ShatterRank
								end

								if spellName == "Ice Lance" and not iceLanceBonus then
									iceLanceBonus = true
									spellDmgM = spellDmgM  * 3
									bDmgM = bDmgM  + 2
								end
							elseif ( buffName == "Moonfire" or buffName == "Insect Swarm" ) then
								if spellName == "Starfire" then
									debuffBonus = true
								end
							end
						else
							BuffCalc( buffName, buffData, apps, texture, rank )
						end
					end
				end
			end
		end
	end
	
	--BUFF -- DAMAGE/HEALING -- PLAYER
	if DrDamage.DamageBuffs and not baseSpell.NoBuffs then
	
		for buffName, index, apps, texture, rank in SEA:BuffIter( "player" ) do
		
			if BS:HasReverseTranslation( buffName ) then
				buffName = BS:GetReverseTranslation( buffName )
			end		
		
			if DrDamage.DamageBuffs[buffName] then
						
				local buffData = DrDamage.DamageBuffs[buffName]
				
				if not healingSpell
					and ( DrDamage:MatchData( buffData.School, spellSchool ) and not baseSpell.NoSchoolBuffs
					or ( not buffData.School and not buffData.Spell ))
					or DrDamage:MatchData( buffData.School, spellType )
				or healingSpell 
					and buffData.School == "Healing"
				or buffData.School == "All" and spellType ~= "Physical"
				or DrDamage:MatchData( buffData.Spell, lSpellName ) then
				
					if buffData.ModType == "Special" then
						if buffName == "Fel Armor" then
							LeechBonus = LeechBonus * 1.2
						elseif buffName == "Master Demonologist" then
							if UnitCreatureFamily( "pet" ) then
								local petFamily = UnitCreatureFamily("pet")
								
								if petFamily == "Succubus" and BuffTalentRank > 0 then
									dmgM = dmgM * ( 1 + 0.02 * BuffTalentRank )				
								elseif petFamily == "Felguard" then
									dmgM = dmgM * ( 1 + 0.01 * BuffTalentRank )
								end
							end
						elseif buffName == "Touch of Shadow" then
							GT:SetUnitBuff( "player", index )
							local _, _, amount = GT:Find( "(%d+)%%" )
							
							if amount == "10" then
								dmgM = dmgM * 1.1
							else
								dmgM = dmgM * 1.15
							end
						elseif buffName == "Totem of Wrath" then
							hitPerc = hitPerc + 3
						elseif buffName == "Vengeance" or buffName == "Revenge" then
							if apps and BuffTalentRank then
								dmgM = dmgM * ( 1 + 0.01 * BuffTalentRank * apps )
							end
						elseif buffName == "Sanctity Aura" then
							if impSanctity then
								if spellSchool == "Holy" then
									dmgM = dmgM * ( 1.1 + 0.01 * impSanctity )
								else
									dmgM = dmgM * ( 1 + 0.01 * impSanctity )
								end
							else
								if spellSchool == "Holy" then
									dmgM = dmgM * 1.1
								end
							end
						end
					else
						BuffCalc( buffName, buffData, apps, texture, rank )
					end				
				end
			end		
		end
	end
	
	--DEBUFF -- DAMAGE/HEALING -- PLAYER
	if DrDamage.DamageDebuffs and not baseSpell.NoBuffs then
	
		for buffName, apps, _, texture, rank in SEA:DebuffIter( "player" ) do
		
			if BS:HasReverseTranslation( buffName ) then
				buffName = BS:GetReverseTranslation( buffName )
			end		
		
			if DrDamage.DamageDebuffs[buffName] then
						
				local buffData = DrDamage.DamageDebuffs[buffName]
				
				if not healingSpell 
					and ( DrDamage:MatchData( buffData.School, spellSchool ) 
					or ( not buffData.School and not buffData.Spell ))
				or healingSpell 
					and buffData.School == "Healing"					
				or DrDamage:MatchData( buffData.Spell, lSpellName ) then	
				
					if buffData.ModType == "Special" then
						if buffName == "Arcane Blast" then
							castTime = castTime - (1/3) * apps
						end
					else
						BuffCalc( buffName, buffData, apps, texture, rank )
					end
				end
			end
		end
	end
	
	--BUFF - HEALING -- TARGET
	if healingSpell and DrDamage.HealingBuffs then
	
		local healingTarget

		if UnitIsFriend( "player", "target" ) then
			healingTarget = "target"
		elseif UnitIsFriend( "player", "targettarget" ) then
			healingTarget = "targettarget" 
		end
		
		if healingTarget then
		
			for buffName, index, apps, texture, rank in SEA:BuffIter( healingTarget ) do

				if BS:HasReverseTranslation( buffName ) then
					buffName = BS:GetReverseTranslation( buffName )
				end		

				if DrDamage.HealingBuffs[buffName] then

					local buffData = DrDamage.HealingBuffs[buffName]

					if not buffData.Spell or DrDamage:MatchData( buffData.Spell, lSpellName ) then

						--if buffData.ModType == "Special" then
						--else
							BuffCalc( buffName, buffData, apps, texture, rank )
						--end
					end
				end
			end
		end
	end
	
	
	--Class specific custom mods
	if englishClass == "PALADIN" and jotcBonus then
	
		if IsEquippedItem( 23203 ) then						--Libram of Fervor
			jotcBonus = jotcBonus + 33
		elseif IsEquippedItem( 27983 ) or IsEquippedItem( 27949 ) then		--Libram of Zeal
			jotcBonus = jotcBonus + 48					
		end

		if BuffTalentRank2 > 0 then
			jotcBonus = jotcBonus * ( 1 + 0.05 * BuffTalentRank2 )
		end

		if BuffTalentRank3 > 0 then
			critPerc = critPerc + BuffTalentRank3
		end
	elseif englishClass == "SHAMAN" then
	
		--Purification doesn't work on earth shield if the target is ~= player
		if BuffTalentRank > 0 and not UnitIsUnit( "target", "player" ) then
			dmgM = dmgM / ( 1 + BuffTalentRank * 0.02 )
		end
	end
	
	--Calculate in the most important set bonuses
	if type( DrDamage.SetBonuses ) == "function" then
		eDuration, finalMod, dmgM, castTime, chainFactor, jotcBonus = DrDamage:SetBonuses( spellName, eDuration, finalMod, dmgM, castTime, chainFactor, jotcBonus, debuffBonus )
	end
	
	--Modified factors?
	
	--Split between bonus to dot and nuke
	if baseSpell.hybridFactor then
		spellDmgM = spellDmgM * baseSpell.hybridFactor
	end
	
	--Factor that applies to all ranks of the spell
	if baseSpell.sFactor then
		spellDmgM = spellDmgM * baseSpell.sFactor
	end
	
	--Factor that applies to current rank of the spell
	if spell.sFactor then
		spellDmgM = spellDmgM * spell.sFactor
	end
	
	
	--Factor for downranking
	local lowRankFactor = 1
	
	if spell.maxLevel and spell.maxLevel < 68 then
		
		local playerLevel = UnitLevel("player")
		
			local downRankMod = 0
			
			if spell.Downrank then
				downRankMod = spell.Downrank
			elseif baseSpell.Downrank then
				downRankMod = baseSpell.Downrank
			end
	
			if healingSpell then
				lowRankFactor = ( spell.maxLevel + downRankMod + 6 ) / playerLevel
			else
				lowRankFactor = ( spell.maxLevel + downRankMod + 4 ) / playerLevel
			end
		
			if lowRankFactor < 1 then
				spellDmgM = spellDmgM * lowRankFactor
			else
				lowRankFactor = 1
			end
	end
	
	--Factor for spells under level 20
	local lowLevelFactor = 1
	
	if spell.spellLevel and spell.spellLevel < 20 and not baseSpell.NoLowLevelPenalty then
	
		lowLevelFactor = 1 - ((20 - spell.spellLevel) * 0.0375)
		
		if lowLevelFactor < 1 and lowLevelFactor > 0 then
			spellDmgM = spellDmgM * lowLevelFactor
		else
			lowLevelFactor = 1
		end
	end
	
	--Adding to spells base damage after levelups:
	if baseSpell.BaseIncrease and spell.spellLevel then
	
		local playerLevel = UnitLevel("player")
	
		if playerLevel > spell.spellLevel then
				
			if spell.maxLevel and playerLevel >= spell.maxLevel then
				minDam = minDam + spell[3]
				maxDam = maxDam + spell[4]
				
			elseif baseSpell.NextRank and playerLevel >= ( spell.spellLevel + baseSpell.NextRank - 1 ) then
				minDam = minDam + spell[3]
				maxDam = maxDam + spell[4]			
			else
				local diff = 1
				
				if baseSpell.NextRank then
					diff = baseSpell.NextRank - 1
				elseif spell.maxLevel then
					diff = spell.maxLevel - spell.spellLevel
				end
				
				minDam = minDam + ( playerLevel - spell.spellLevel ) * ( spell[3] / diff )
				maxDam = maxDam + ( playerLevel - spell.spellLevel ) * ( spell[4] / diff )
			end
		end
		
		minDam = DrDamage:Round( minDam )
		maxDam = DrDamage:Round( maxDam )
	end
	
	--Special cases:
	if englishClass == "PALADIN" then
		if spellName == "Crusader Strike" then
			
			local baseMin, baseMax = DrDamage:BaseMelee()
			local atkPower = DrDamage:TotalAP()
			local normM
			
			if IsEquippedItemType( "One-Handed Axes" ) or IsEquippedItemType( "One-Handed Maces" ) or IsEquippedItemType( "One-Handed Swords"  ) then
				normM = 2.4
			else
				normM = 3.3
			end
			
			minDam = baseMin + ( atkPower / 14 ) * normM
			maxDam = baseMax + ( atkPower / 14 ) * normM
		end
		
		if bolBonus then
			if spellName == "Holy Light" or spellName == "Flash of Light" then
				if IsEquippedItem( 28592 ) then		--Libram of Souls Redeemed
					finalMod = finalMod + 105
				end
			end
		end
	end
	
	--Adds JotC to holy spells
	if jotcBonus and spellSchool == "Holy" then
		spellDmg = spellDmg + jotcBonus
	end	
	
	
	--Soul siphon
	if afflCount and SSRank then

		local SSBonus = 0

		if SSRank == 1 then
			SSBonus = afflCount * 0.02

			if SSBonus > 0.24 then
				SSBonus = 0.24
			end
		elseif SSRank == 2 then
			SSBonus = afflCount * 0.05

			if SSBonus > 0.60 then
				SSBonus = 0.60
			end
		end

		dmgM = dmgM * ( 1 + SSBonus )	
	end
	
	local function DmgCalc( next, hitCalc )
	
		--Initialize variables that are calculated
		local dispSpellDmgM = spellDmgM		--Calculated spelldamage modifier		
		local calcAvgDmgCrit = 0		--Calculated final damage with averaged crit
		local calcAvgDmg = 0			--Calculated final damage without crit
		local calcDotDmg = 0			--Calculated DOT portion
		local calcMinDmg = 0			--Calculated dmg range: min
		local calcMaxDmg = 0			--Calculated dmg range: max
		local calcMinCrit			--Calculated crit range: min
		local calcMaxCrit			--Calculated crit range: max
		local igniteDotDmg			--Calculated ignite DOT portion
		local avgIgnite				--Calculated average ignite
		local perHit				--Calculated damage per hit
		local extraDmg				
		local extraDPS
		
		if sHits then
			calcMinDmg = dmgM * ( bDmgM * minDam * sHits + (  spellDmg * spellDmgM ) )
			calcMaxDmg = dmgM * ( bDmgM * maxDam * sHits + (  spellDmg * spellDmgM ) )
		elseif spell.hybridDotDmg then
			calcMinDmg = ( dmgM + bNukeDmg ) * ( bDmgM * minDam + ( spellDmg * spellDmgM ) )
			calcMaxDmg = ( dmgM + bNukeDmg ) * ( bDmgM * maxDam + ( spellDmg * spellDmgM ) )
		else
			calcMinDmg = dmgM * ( bDmgM * minDam + ( spellDmg * spellDmgM ) )
			calcMaxDmg = dmgM * ( bDmgM * maxDam + ( spellDmg * spellDmgM ) )
		end
				
		if finalMod ~= 0 then
		
			if finalMod > 0 then
				calcMinDmg = calcMinDmg + finalMod * dmgM
				calcMaxDmg = calcMaxDmg + finalMod * dmgM
			else
				calcMinDmg = calcMinDmg + finalMod
				calcMaxDmg = calcMaxDmg + finalMod
			end
				
			if calcMinDmg < 0 then
				calcMinDmg = 0
			end
			
			if calcMaxDmg < 0 then
				calcMaxDmg = 0
			end	
		end				
		
		calcAvgDmg = (calcMinDmg + calcMaxDmg) / 2
				
		--This is for Imp SW:P and others with extended duration through talents etc.
		if baseSpell.eDot and baseSpell.eDuration and eDuration > baseSpell.eDuration then
			calcAvgDmg = calcAvgDmg + ( calcAvgDmg / baseSpell.eDuration ) * ( eDuration - baseSpell.eDuration )
		end		
		
		--Crit calculation:
		local critBonus = 0
		
		if critPerc > 100 then
			critPerc = 100
		end
		
		if baseSpell.canCrit then
			calcMinCrit = calcMinDmg + calcMinDmg * critM
			calcMaxCrit = calcMaxDmg + calcMaxDmg * critM
			critBonus = ( critPerc / 100 ) * calcAvgDmg * critM
			calcAvgDmgCrit = calcAvgDmg + critBonus
		else
			calcAvgDmgCrit = calcAvgDmg
			critPerc = 0;
		end

		--Hit calculation:
		if hitPerc > 99 then
			hitPerc = 99
		end
		
		if not healingSpell and not baseSpell.Unresistable then
			if DrDamage.db.profile.HitCalc or hitCalc then
				if DrDamage.db.profile.TwoRoll then
					calcAvgDmgCrit = calcAvgDmgCrit * ( hitPerc / 100 )
				else
					calcAvgDmgCrit = ( calcAvgDmgCrit - critBonus ) * ( hitPerc / 100 ) + critBonus
				end
			end
		end		
		
		--Hybrid spells:
		local hybridDmgM = 0
		
		if spell.hybridDotDmg then
	
			local hybridFactor = 0
			local sFactor = 1
			
			if baseSpell.hybridFactor then
				hybridFactor = baseSpell.hybridFactor
			end

			if baseSpell.sFactor then
				sFactor = baseSpell.sFactor * sFactor
			end

			if spell.sFactor then
				sFactor = spell.sFactor * sFactor
			end

			if baseSpell.hybridDotFactor then
				sFactor = baseSpell.hybridDotFactor * sFactor
			end
			
			local baseDuration = spell.eDuration or baseSpell.eDuration
			
			hybridDmgM = ( baseDuration / 15 + dotSpellDmgM ) * ( 1 - hybridFactor ) * sFactor * lowRankFactor * lowLevelFactor
			calcDotDmg = dmgM * ( spell.hybridDotDmg + hybridDmgM * spellDmg )
			dispSpellDmgM = dispSpellDmgM + hybridDmgM
		
			if eDuration > baseDuration then 
				calcDotDmg = calcDotDmg + ( calcDotDmg / baseDuration ) * ( eDuration - baseDuration )
			end
		end
		
		--For spells that can ignite
		if baseSpell.Ignite then
			igniteDotDmg = igniteM * (( calcMinCrit + calcMaxCrit ) / 2 )
			avgIgnite = ( critPerc / 100 ) * igniteDotDmg
		end
		
		--For spells with extra DOT portion eg. fireball
		if spell.extraDotDmg then
		
			local xFactor = 0
			
			if baseSpell.extraDotF then
				xFactor = baseSpell.extraDotF
			end
			
			calcDotDmg = calcDotDmg + dmgM * ( spell.extraDotDmg + xFactor * lowRankFactor * lowLevelFactor * spellDmg ) --TODO: Verify factor
			dispSpellDmgM = dispSpellDmgM + xFactor * lowRankFactor * lowLevelFactor
		end
		
		calcDotDmg = calcDotDmg + dotFinalMod
		
		--For chain effects (chain lightning, chain heal)
		local chainEffect
		
		if chainFactor > 0 then
			chainEffect = calcAvgDmgCrit * ( chainFactor + ( chainFactor ^ 2 ) + ( chainFactor ^ 3 ) )
			calcAvgDmgCrit = calcAvgDmgCrit + chainEffect
		end
					
		--DPS calculations
		local calcDPS
		
		if castTime < 1.5 and baseSpell.Instant then
			castTime = 1.5
		end
	
		if baseSpell.eDot then
			calcDPS = calcAvgDmgCrit / eDuration
		else
			if baseSpell.Ignite then
			
				calcDPS = calcAvgDmgCrit / castTime + avgIgnite / 4
				
				if spell.hybridDotDmg or spell.extraDotDmg then
					calcDPS = calcDPS + calcDotDmg / eDuration
				end
				
				calcAvgDmgCrit = calcAvgDmgCrit + avgIgnite + calcDotDmg
				
			elseif spell.hybridDotDmg or spell.extraDotDmg then
				calcDPS = calcAvgDmgCrit / castTime + calcDotDmg / eDuration
				calcAvgDmgCrit = calcAvgDmgCrit + calcDotDmg
			else
				calcDPS = calcAvgDmgCrit / castTime
			end
		end
		
		--Special cases:
		if spellName == "Seed of Corruption" then
			calcAvgDmgCrit = calcAvgDmgCrit - calcDotDmg + math_ceil( ( 6 * 1044 ) / calcDotDmg ) * ( calcDotDmg / 6 )
			perHit = calcDotDmg / 6
			calcDotDmg = math_ceil( ( 6 * 1044 ) / calcDotDmg ) * ( calcDotDmg / 6 )
		end
		
		if not next then
		
			local DPS = 			DrDamage:Round( calcDPS, 1 )
			local DPSC = 			DrDamage:Round( calcAvgDmgCrit / castTime, 1 )
			local Avg =			DrDamage:Round( calcAvgDmg )
			local MinDmg = 			math_floor( calcMinDmg )
			local MaxDmg = 			math_ceil( calcMaxDmg )
			
			--Write tooltip data
			if tooltip then

				if baseSpell.Cooldown then
					CalculationResults.DPSCD = 	DrDamage:Round( calcAvgDmgCrit / cooldown, 1 )
				end
				
				if baseSpell.canCrit then
					CalculationResults.MinCrit =	math_floor( calcMinCrit )
					CalculationResults.MaxCrit = 	math_ceil( calcMaxCrit )
					CalculationResults.AvgCrit = 	DrDamage:Round( ( calcMinCrit + calcMaxCrit ) / 2 )
					CalculationResults.CritRate = 	DrDamage:Round( critPerc, 2 )
					
					if not healingSpell and IsEquippedItem( 28785 ) and not DrDamage:MatchData( baseSpell.School, "OffensiveTotem", "Physical" ) then
						extraDmg = ( 700 * ( 1 + ( DrDamage:SpellCrit( "Nature" ) / 100 ) * 0.5 ) ) / ( 300 / critPerc )
						
						if sHits then
							extraDmg = sHits * extraDmg
						end
						
						extraDPS = extraDmg / castTime
					end
				end
				
				CalculationResults.HitRate = 	DrDamage:Round( hitPerc, 2 )
				CalculationResults.DotDmg = 	DrDamage:Round( calcDotDmg )
				CalculationResults.SpellDmg = 	DrDamage:Round( spellDmg )
				CalculationResults.SpellDmgM = 	DrDamage:Round( dispSpellDmgM, 3 )
				CalculationResults.DmgM = 	DrDamage:Round( dmgM, 3 )
				CalculationResults.IgniteDmg = 	DrDamage:Round( igniteDotDmg )
				
				CalculationResults.ChainDmg = 	DrDamage:Round( chainEffect )
				CalculationResults.CastTime = 	castTime
				CalculationResults.Cooldown = 	cooldown
				CalculationResults.ManaRegen = 	DrDamage:Round( manaRegenCasting, 1 )
				CalculationResults.FreeCrit = 	freeCrit
				
				CalculationResults.ExtraDPS = 	DrDamage:Round( extraDPS, 1 )
			
				local sTicks = baseSpell.sTicks or spell.sTicks
				local ticks

				if sHits then
					ticks = sHits
					perHit = calcAvgDmg / sHits
				elseif sTicks then
				
					if castTime > eDuration then
						ticks = castTime / sTicks
					else
						ticks = eDuration / sTicks
					end
					
					if spellName == "Seed of Corruption" then
						ticks = math_ceil( 1044 / perHit )
						CalculationResults.SpellDmgM = DrDamage:Round( dispSpellDmgM -  ( hybridDmgM - ( hybridDmgM / 6 ) * ticks ), 3 )
					else
						if spell.hybridDotDmg or spell.extraDotDmg then
							perHit = calcDotDmg / ticks
						else
							perHit = calcAvgDmg / ticks
						end
					end
				end		
				
				if baseSpell.Leech then
					CalculationResults.AvgLeech = DrDamage:Round( calcAvgDmgCrit * LeechBonus, 1 )
				end
				
				if perHit and ticks then
				
					CalculationResults.PerHit = DrDamage:Round( perHit, 1 )
					CalculationResults.Hits = ticks
					
					if baseSpell.Leech then
						CalculationResults.PerHitHeal =  DrDamage:Round( perHit * LeechBonus, 1 )
					end
				end
			end
					
			return calcAvgDmgCrit, DPS, DPSC, Avg, MinDmg, MaxDmg
		end
				
		return calcAvgDmgCrit
	end
	
	local returnAvgTotal, returnDPS, returnDPSC, returnAvg, returnMinDmg, returnMaxDmg = DmgCalc()
	
	
	if tooltip then
	
		if DrDamage.db.profile.Next then
		
			if baseSpell.canCrit then
				critPerc = critPerc + 1
				CalculationResults.NextCrit = DrDamage:Round( DmgCalc( true ) - returnAvgTotal, 1 )
				critPerc = critPerc - 1	
			end

			spellDmg = spellDmg + 10
			CalculationResults.NextTenDmg = DrDamage:Round( DmgCalc( true ) - returnAvgTotal, 1 )
			spellDmg = spellDmg - 10

			if not healingSpell and not baseSpell.Unresistable then
			
				local temp
				
				if DrDamage.db.profile.HitCalc then
					temp = returnAvgTotal
				else
					temp = DmgCalc( true, true )
				end
				
				hitPerc = hitPerc + 1
				CalculationResults.NextHit = DrDamage:Round( DmgCalc( true, true ) - temp, 1 )
				hitPerc = hitPerc - 1
			end	
		end		
	end
	
	return healingSpell, DrDamage:Round( returnAvgTotal ), returnDPS, returnDPSC, returnAvg, returnMinDmg, returnMaxDmg, CalculationResults
end

function DrDamage:Round(x, y)

	if type( x ) ~= "number" then
		return x
	end

	y = y or 0
	
	x = x * 10 ^ y
	
	if (x > 0) then
		x = math_floor(x+0.5)
  	else
    		x = math_ceil(x-0.5)
  	end
  	
  	x = x / 10 ^ y
  	
  	return x
end

